country_acronyms = [
    "USA",  # United States of America
    "CHN",  # China
    "HKG",  # Hong Kong
    "IND",  # India
    "JPN",  # Japan
    "RUS",  # Russia
    "SAU",  # Saudi Arabia
    "SIN",  # Singapore
    "KOR",  # South Korea
    "SWE",  # Sweden
    "CHE",  # Switzerland
    "ROC",  # Taiwan
    "UKR",  # Ukraine
    "DEU",  #
    "FRA",  # France
    "ARG",  # Argentina  
    "ITA",  # Italy
    "NZL",  # New Zealand
    "AUS",  # Australia
    "CAN",  # Canada
    "EUR",  # Europe

]

repo_codes = [
    "RRPTSYD",  # Treasury Securities Sold by the Federal Reserve
    "RPTSYD",  # Treasury Securities Purchased by the Federal Reserve
    "RPTTLD",  # Total Securities Purchased by the Federal Reserve
    "RPTTLD",  # Overnight Repurchase Agreements: Mortgage Backed Securities Sold
    "RPONMBSD",  # Overnight Repurchase Agreements: Mortgage-Backed Securities Purchased
    "RPONTSYD",  # Overnight Repurchase Agreements: Treasury Securities Purchased
    "RPONAGYD",  # Overnight Repurchase Agreements: Federal Agency Securities Purchased
    "RPMBSD",  # Repurchase Agreements: Mortgage-Backed Securities Purchased
    "RPAGYD",  # Repurchase Agreements: Federal Agency Securities Sold
    "RRPTTLD",  # Repurchase Agreements: Federal Agency Securities Purchased
    "RRPONTTLD",  # Reverse Repurchase Agreements: Total Securities Sold
]

sp500_codes = [
    "SP500_DIV_YIELD_MONTH",  # S&P 500 Dividend Yield by Month
    "SP500_PE_RATIO_MONTH",  # Monthly P/E Ratio
    "SHILLER_PE_RATIO_YEAR",  # Shiller PE Ratio by Year
    "SP500_PBV_RATIO_QUARTER",  # S&P 500 Price to Book Value by Quarter
    "SP500_REAL_PRICE_MONTH",  # S&P 500 Real Price by Month
    "SP500_INFLADJ_MONTH",  # S&P 500 Inflation Adjusted by Month
    "SP500_INFLADJ_YEAR",  # S&P 500 Inflation Adjusted by Year
    "SP500_BVPS_QUARTER",  # S&P 500 Book Value Per Share by Quarter
    "SP500_EARNINGS_GROWTH_QUARTER",  # S&P 500 Earnings Growth Rate by Quarter
    "SP500_SALES_QUARTER"  # S&P 500 Sales by Quarter
]


income_expenditure_codes = [
    "MEHOINUSA672N",  # Real Median Household Income in the United States
    "DSPIC96",  # Real Disposable Personal Income
    "PCE",  # Personal Consumption Expenditures
    "PCEDG",  # Personal Consumption Expenditures: Durable Goods
    "PSAVERT",  # Personal Saving Rate
    "RRSFS",  # Real Retail and Food Services Sales
    "DSPI"  # Disposable Personal Income
]


economic_indicator_codes = [ 

 
    "INDPRO",  # Industrial Production Index
    "TCU",  # Capacity Utilization: Total Industry
    "HOUST",  # Housing Starts: Total: New Privately Owned Housing Units Started
    "GPDI",  # Gross Private Domestic Investment
    "cp",  # Corporate Profits After Tax (without IVA and CCAdj)
    "STLFSI",  # St. Louis Fed Financial Stress Index
    "DCOILWTICO",  # Crude Oil Prices: West Texas Intermediate (WTI)
    "USSLIND",  # Leading Index for the United States
    "DTWEXM",  # Trade Weighted U.S. Dollar Index: Major Currencies
    "DTWEXB"  # Trade Weighted U.S. Dollar Index: Broad

]

unemployment_codes = [
    "UNRATE",  # Civilian Unemployment Rate
    "NROU",  # Natural Rate of Unemployment (Long-Term)
    "NROUST",  # Natural Rate of Unemployment (Short-Term)
    "CIVPART",  # Civilian Labor Force Participation Rate
    "EMRATIO",  # Civilian Employment-Population Ratio
    "UNEMPLOY",  # Unemployed
    "PAYEMS",  # All Employees: Total nonfarm
    "MANEMP",  # All Employees: Manufacturing
    "ICSA",  # Initial Claims
    "IC4WSA"  # 4-Week Moving Average of Initial Claims
]


interest_rate_codes = [
    "DFF",  # Effective Federal Funds Rate
    "DTB3",  # 3-Month Treasury Bill: Secondary Market Rate
    "DGS5",  # 5-Year Treasury Constant Maturity Rate
    "DGS10",  # 10-Year Treasury Constant Maturity Rate
    "DGS30",  # 30-Year Treasury Constant Maturity Rate
    "T5YIE",  # 5-year Breakeven Inflation Rate
    "T10YIE",  # 10-year Breakeven Inflation Rate
    "T5YIFR",  # 5-Year, 5-Year Forward Inflation Expectation Rate
    "TEDRATE",  # TED Spread
    "DPRIME"  # Bank Prime Loan Rate
]

gdp_codes = [
    "GDP",  # Gross Domestic Product
    "GDPC1", # Real Gross Domestic Product
    "GDPPOT" 
]


import sys
from pathlib import Path
import json
# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)
import os
from dotenv import load_dotenv
from apis.discord_.discord_sdk import DiscordSDK
from market_handlers.database_ import MarketDBManager
load_dotenv()
from polygon.websocket import WebSocketClient, Market
from polygon.websocket.models import WebSocketMessage, EquityAgg,EquityQuote,EquityTrade,IndexValue
from fudstop.apis.polygonio.mapping import option_condition_desc_dict,option_condition_dict,OPTIONS_EXCHANGES,stock_condition_desc_dict,stock_condition_dict,indicators,quote_conditions,STOCK_EXCHANGES
from list_sets.dicts import all_forex_pairs, crypto_currency_pairs
# Create a reverse dictionary
all_forex_pairs = {v: k for k, v in all_forex_pairs.items()}
from typing import List
import asyncio
from asyncio import Queue
import pandas as pd
import logging
class MultiMarkets:
    def __init__(self):

        self.discord = DiscordSDK()
        self.db = MarketDBManager(user='chuck', database='fudstop')
        self.markets = [Market.Options, Market.Stocks, Market.Crypto, Market.Forex, Market.Indices]
        self.subscription_patterns = {
            Market.Options: ["T.*,A.*"],
            Market.Stocks: ["A.*,T.*,Q.*"],
            Market.Indices: ["A.*"],
            Market.Crypto: ['XAS.*, XT.*'],
            Market.Forex: ['CAS.*']

        }


        self.created_channels = set()  # A set to keep track of created channels

    async def create_channel_if_not_exists(self, ticker, name):
        # Check if the channel already exists
        if ticker not in self.created_channels:
            # If not, create the channel and add its name to the set
            await self.discord.create_channel(name=ticker, channel_description=name)
            self.created_channels.add(ticker)










    # Function to handle incoming WebSocket messages
    async def handle_msg(self, msgs: WebSocketMessage):
        
        for m in msgs:
            
            event_type = m.event_type
            if event_type == 'A' and m.symbol.startswith('I:'):
                async for m in self.db.insert_indices_aggs(m):
                    pass


            if event_type == 'A' and m.symbol.startswith('O:'):
                async for m in self.db.insert_option_aggs(m):
                    pass
        


            if event_type == 'A' and not m.symbol.startswith('I:') and not m.symbol.startswith("O:"):
                async for m in self.db.insert_stock_aggs(m):
                    print(m)





            if event_type == 'T' and m.symbol.startswith('O:'):
                async for m in self.db.insert_option_trades(m):
                    pass


            if event_type == 'T' and not m.symbol.startswith('O:') and not m.symbol.startswith('I:'):
                async for m in  self.db.insert_stock_trades(m):
                    pass






            if event_type == 'Q':
                async for m in self.db.insert_stock_quotes(m):
                    pass

            if event_type == 'XT':
                async for data in self.db.insert_crypto_trades(m):
                    
                    print(data)

                    

            if event_type == 'CAS':
                async for data in self.db.insert_forex_aggs(m=m):

                    name = all_forex_pairs.get(data['pair'])
                    

                    await self.discord.create_channel(guild_id=self.discord.fudstop_id, name=name, type='0', channel_description=f"Live forex from polygon.io! Go to https://www.polygon.io and use code FUDSTOP  at checkout for a 10% discount.> Forex Pair Monitored: {name} : Symbol: {data['pair']}")


market = MultiMarkets()


async def main():
    await market.db.connect()
    while True:  # Restart mechanism
        try:
            await run_main_tasks()
        except Exception as e:
            print(e)
            logging.error(f"Critical error in main loop: {e}")
            logging.info("Restarting main loop...")
            await asyncio.sleep(10)  # Pause before restarting

# Main async function to connect to all markets with their respective subscriptions
async def run_main_tasks():

    clients = []
    for live_market in market.markets:
        patterns = market.subscription_patterns.get(live_market, [])
        for pattern in patterns:
            client = WebSocketClient(subscriptions=[pattern], api_key=os.environ.get('YOUR_POLYGON_KEY'), market=live_market)
            clients.append(client.connect(market.handle_msg))
    await asyncio.gather(*clients)

# Run the main function
asyncio.run(main())



# At the top of the file.
import disnake
from disnake.ext import commands
from disnake import TextInputStyle
import aiohttp
import os
import asyncio
import openai
# Subclassing the modal.
class OptionModalView(disnake.ui.View):
    def __init__(self):
        super().__init__(timeout=None)



        self.add_item(QuoteSelect())
        self.add_item(GreekSelect())
        self.add_item(TradeSelect())
        self.add_item(MarketDataSelect())
        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji=f"<a:_:1044503166433099856>", row=4)
        self.add_item(button)



class GreekSelect(disnake.ui.Select):

    def __init__(self):
        super().__init__( 
            placeholder=f"Filter Greeks -->",
            custom_id='filtergreeks',
            min_values=1,
            max_values=5,
            options = [
                    disnake.SelectOption(
                        label="Gamma",
                        description="Measures Delta's change per $1 increase in stock."
                    ),
                    disnake.SelectOption(
                        label="Vega",
                        description="Option price sensitivity to volatility."
                    ),
                    disnake.SelectOption(
                        label="Delta",
                        description="Option price change per $1 move in stock."
                    ),
                    disnake.SelectOption(
                        label="IV",
                        description="Estimate of future security price volatility."
                    ),
                    disnake.SelectOption(
                        label="Theta",
                        description="Option price time decay rate per day."
                    )
])

    
    async def callback(self, interaction: disnake.MessageCommandInteraction):
        # Assuming you have a function to create and return a modal
        modal = self.create_modal_based_on_selection(self.values)
        await interaction.response.send_modal(modal)

    def create_modal_based_on_selection(self, selections):
        components = []
        
        for selection in selections:
            components.append(
            disnake.ui.TextInput(
                label=selection,
                custom_id=f'input_{selection.lower()}',
                style=disnake.TextInputStyle.short,
                placeholder=f'Enter {selection} value...',
            
            ))

        
        return GreeksModal(components)





class OptionsDataModal(disnake.ui.Modal):
    """
    Perform options related tasks
    
    """
    def __init__(self):
        # The details of the modal, and its components
        components = [
            disnake.ui.TextInput(
                label="Underlying Symbol",
                placeholder="e.g. AAPL",
                custom_id="symbol",
                style=TextInputStyle.paragraph,
                max_length=8,
            ),
            disnake.ui.TextInput(
                label="Strike Price",
                placeholder="e.g. 12.5 | 30",
                custom_id="strikeprice",
                style=TextInputStyle.paragraph,
                max_length=8,
            ),
            disnake.ui.TextInput(
                label="Call or Put?",
                placeholder="e.g. call | put",
                custom_id="callput",
                style=TextInputStyle.paragraph,
                max_length=4,
            ),

            disnake.ui.TextInput(
                label="Expiry Date",
                placeholder="e.g. 2023-12-15",
                custom_id="expiry",
                style=TextInputStyle.paragraph,
                max_length=10,
            ),

        ]
        super().__init__(title="Build your Symbol:", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
        await inter.response.defer(with_message=True, ephemeral=False)

        embed = disnake.Embed(title=f"Option Loaded - {inter.text_values.get('symbol')} | ${inter.text_values.get('strikeprice')} | {inter.text_values.get('callput')} | {inter.text_values.get('expiry')}", description=f"> # **Use the dropdown below to choose data**", color=disnake.Colour.dark_teal())
        await inter.edit_original_message(embed=embed, view=OptionModalView())


class GreeksModal(disnake.ui.Modal):
    """
    Perform options related tasks
    
    """
    def __init__(self, components):
        components = components
        # The details of the modal, and its components
        # components = [
        #     disnake.ui.TextInput(
        #         label="Enter a delta range -->",
        #         placeholder="e.g. delta >= 0.5",
        #         custom_id="delta",
        #         style=TextInputStyle.short,
        #         max_length=100,
        #         required=False
        #     ),
        #     disnake.ui.TextInput(
        #         label="Enter a gamma range -->",
        #         placeholder="e.g. >= 0.07",
        #         custom_id="gamma",
        #         style=TextInputStyle.short,
        #         max_length=100,
        #         required=False
        #     ),
        #     disnake.ui.TextInput(
        #         label="Enter a theta range -->",
        #         placeholder="e.g. <= 0.02",
        #         custom_id="theta",
        #         style=TextInputStyle.short,
        #         max_length=10,
        #         required=False
        #     ),

        #     disnake.ui.TextInput(
        #         label="Enter a vega range -->",
        #         placeholder='e.g. <= 0.05',
        #         custom_id="vega",
        #         style=TextInputStyle.short,
        #         max_length=10,
        #         required=False
        #     ),

        #     disnake.ui.TextInput(
        #         label="Enter an IV range..",
        #         placeholder="e.g. >= 25",
        #         custom_id="iv",
        #         style=TextInputStyle.short,
        #         max_length=7,
        #     ),

        # ]
        super().__init__(title="Filter By Greeks:", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
        await inter.response.defer(with_message=True)

        await inter.edit_original_message('> # Chain Modal worked! Prepare autism!')



class QuoteSelect(disnake.ui.Select):

    def __init__(self):
        options = [
            disnake.SelectOption(
                label="Ask",
                description="Enter the asking price."
            ),
            disnake.SelectOption(
                label="Ask Size",
                description="Enter the size of the ask."
            ),
            disnake.SelectOption(
                label="Bid",
                description="Enter the bidding price."
            ),
            disnake.SelectOption(
                label="Bid Size",
                description="Enter the size of the bid."
            ),
            disnake.SelectOption(
                label="Midpoint",
                description="Enter the midpoint value."
            )
        ]
        super().__init__(
            placeholder="Filter Quotes -->",
            custom_id='filterquotes',
            min_values=1,
            max_values=5,
            options=options
        )
    
    async def callback(self, interaction: disnake.MessageCommandInteraction):
        modal = self.create_modal_based_on_selection(self.values)
        await interaction.response.send_modal(modal)

    def create_modal_based_on_selection(self, selections):
        components = []
        
        for selection in selections:
            components.append(
                disnake.ui.TextInput(
                    label=selection,
                    custom_id=f'input_{selection.lower().replace(" ", "_")}',
                    style=disnake.TextInputStyle.short,
                    placeholder=f'Enter {selection} value...',
                )
            )
        
        return QuotesModal(components)


class QuotesModal(disnake.ui.Modal):
    """
    Perform options related tasks
    
    """
    def __init__(self, components):
        components = components


        super().__init__(title="Filter the quotes:", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
        await inter.response.defer(with_message=True)

        await inter.edit_original_message('> # Quotes chain - Worked!')



class TradeSelect(disnake.ui.Select):

    def __init__(self):
        options = [
            disnake.SelectOption(
                label="Conditions",
                description="Trade conditions."
            ),
            disnake.SelectOption(
                label="Exchange",
                description="Exchange identifier."
            ),
            disnake.SelectOption(
                label="Price",
                description="Trade price."
            ),
            disnake.SelectOption(
                label="Timestamp",
                description="SIP timestamp."
            ),
            disnake.SelectOption(
                label="Size",
                description="Trade size."
            )
        ]
        super().__init__(
            placeholder="Filter Trades -->",
            custom_id='filtertrades',
            min_values=1,
            max_values=5,
            options=options
        )


    async def callback(self, interaction: disnake.MessageCommandInteraction):
            modal = self.create_modal_based_on_selection(self.values)
            await interaction.response.send_modal(modal)

    def create_modal_based_on_selection(self, selections):
        components = []
        
        for selection in selections:
            components.append(
                disnake.ui.TextInput(
                    label=selection,
                    custom_id=f'input_{selection.lower().replace(" ", "_")}',
                    style=disnake.TextInputStyle.short,
                    placeholder=f'Enter {selection} value...',
                )
            )
        
        return TradeModal(components)




class TradeModal(disnake.ui.Modal):
    """
    Perform options related tasks
    
    """
    def __init__(self, components):
        components = components


        super().__init__(title="Filter the trades:", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
        await inter.response.defer(with_message=True)

        await inter.edit_original_message('> # Trades chain - Worked!')



class MarketDataSelect(disnake.ui.Select):

    def __init__(self):
        options = [
            disnake.SelectOption(
                label="Change Percent",
                description="Change in price by percentage."
            ),
            disnake.SelectOption(
                label="Open Interest",
                description="Total outstanding contracts."
            ),
            disnake.SelectOption(
                label="Volume",
                description="Volume of trades."
            ),
            disnake.SelectOption(
                label="VWAP",
                description="Volume Weighted Average Price."
            ),
            disnake.SelectOption(
                label="Underlying Price",
                description="Price of the underlying asset."
            )
        ]
        super().__init__(
            placeholder="Select Market Data -->",
            custom_id='filtermarketdata',
            min_values=1,
            max_values=5,
            options=options
        )

    async def callback(self, interaction: disnake.MessageCommandInteraction):
        modal = self.create_modal_based_on_selection(self.values)
        await interaction.response.send_modal(modal)

    def create_modal_based_on_selection(self, selections):
        components = []
        
        for selection in selections:
            component_custom_id = f'input_{selection.lower().replace(" ", "_")}'
            components.append(
                disnake.ui.TextInput(
                    label=selection,
                    custom_id=component_custom_id,
                    style=disnake.TextInputStyle.short,
                    placeholder=f'Enter {selection}...',
                )
            )
        
        return MarketDataModal(components)
    



class MarketDataModal(disnake.ui.Modal):
    """
    Perform options related tasks
    
    """
    def __init__(self, components):
        components = components


        super().__init__(title="Filter Acivity:", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
        await inter.response.defer(with_message=True)

        await inter.edit_original_message('> # Activity chain - Worked!')


from apis.polygonio.polygon_database import PolygonDatabase
from polygon.websocket import WebSocketMessage
from apis.helpers import convert_to_ns_datetime
import os
import json
from datetime import datetime
from pytz import timezone
from apis.polygonio.mapping import stock_condition_desc_dict,stock_condition_dict,STOCK_EXCHANGES,OPTIONS_EXCHANGES,option_condition_desc_dict,option_condition_dict,indicators,quote_conditions
import pandas as pd
utc = timezone('UTC')
aware_datetime = utc.localize(datetime.utcnow())
from .list_sets import crypto_conditions_dict, crypto_exchanges
class MarketDBManager(PolygonDatabase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


  

    async def insert_stock_trades(self, m):
        m.condition = stock_condition_dict.get(m.conditions)
        df = pd.DataFrame(vars(m), index=[0])





        await self.batch_insert_dataframe(df, table_name='stock_trades', unique_columns='insertion_timestamp')
        yield m

    async def insert_crypto_trades(self, m):
        



        conditions = [crypto_conditions_dict.get(i) for i in m.conditions]
        data = { 
            'type': m.event_type,
            'symbol': m.pair,
            'exchange': crypto_exchanges.get(m.exchange),
            'id': m.id,
            'price': m.price,
            'size': m.size,
            'conditions': conditions[0],
        }

    
        df = pd.DataFrame(data, index=[0])
        await self.batch_insert_dataframe(df, table_name='crypto_trades', unique_columns='insertion_timestamp')
        yield data
    async def insert_forex_aggs(self, m):
        data_quotes= { 
            'open': m.open,
            'close': m.close,
            'pair': m.pair,
            'high': m.high,
            'low': m.low,
            'volume': m.volume

        }
        df = pd.DataFrame(data_quotes, index=[0])
        await self.batch_insert_dataframe(df, table_name='forex_aggs', unique_columns='insertion_timestamp')

        yield data_quotes


    async def insert_stock_aggs(self, m):
        
        df = pd.DataFrame(vars(m), index=[0])
        await self.batch_insert_dataframe(df, table_name='stock_aggs', unique_columns='insertion_timestamp')
        yield m

    async def insert_stock_quotes(self, m):
        
        m.indicators = ','.join(map(str, m.indicators)) if m.indicators is not None else None
        m.indicators = indicators.get(m.indicators)
        df = pd.DataFrame(vars(m), index=[0])
        await self.batch_insert_dataframe(df, table_name='stock_quotes', unique_columns='insertion_timestamp')
        yield m
    async def insert_option_aggs(self, m):

        
        df = pd.DataFrame(vars(m), index=[0])
        await self.batch_insert_dataframe(df, table_name='option_aggs', unique_columns='insertion_timestamp')
        yield m

    async def insert_option_trades(self, m):
        

        m.conditions = ','.join(map(str, m.conditions))
        m.condition = option_condition_dict.get(m.conditions)
        df = pd.DataFrame(vars(m), index=[0])
        await self.batch_insert_dataframe(df, table_name='option_trades', unique_columns='insertion_timestamp')

        yield m
    async def insert_indices_aggs(self, m):
        

        data_queue_data = { 
            'type': 'Indices',
            'ticker': m.symbol,
            'day_open': m.official_open_price,
            'minute_open': m.open,
            'minute_high': m.high,
            'minute_low': m.low,
            'minute_close': m.close,
            'minute_start': convert_to_ns_datetime(m.start_timestamp),
            'minute_end': convert_to_ns_datetime(m.end_timestamp)
        }


        df = pd.DataFrame(data_queue_data, index=[0])

        await self.batch_insert_dataframe(df, table_name='indices_aggs', unique_columns='insertion_timestamp')      

        
        yield data_queue_data







import disnake
import requests
from disnake.ext import commands

from discord_.bot_menus.pagination import AlertMenus
from apis.helpers import format_large_number
import pandas as pd

class Accounting(commands.Cog):
    def __init__(self, bot):
        self.bot=bot
     





    @commands.slash_command()
    async def accounting(self, inter):
        pass






    @accounting.sub_command()
    async def weekly_treasury_offerings(self, inter: disnake.AppCmdInter):
        """🖋️ Details of Offerings for Regular Weekly Treasury Bills"""

        await inter.response.defer()
        

        r = requests.get("https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/tb/pdo1_offerings_regular_weekly_treasury_bills?sort=-record_date&limit=1000").json()
        data = r['data']

        df = pd.DataFrame(data)
        df.to_csv('data/weekly_treasury_offerings.csv')
        embeds = []

        for i, row in df.iterrows():
            record_date = row["record_date"]
            issue_date = row["issue_date"]
            maturity_date = row["maturity_date"]
            days_to_maturity = row["days_to_maturity"]
            bids_tendered = float(row["bids_tendered_mil_amt"]) * 1000000
            bids_acc_total = float(row["bids_acc_total_mil_amt"]) * 1000000
            bids_acc_comp_basis = float(row["bids_acc_comp_basis_mil_amt"]) * 1000000
            bids_acc_noncomp_basis = float(row["bids_acc_noncomp_basis_mil_amt"]) * 1000000
            high_price_per_hundred = row["high_price_per_hundred"]
            high_discount_rate = row["high_discount_rate"]
            high_investment_rate = row["high_investment_rate"]
            src_line_nbr = row["src_line_nbr"]
            record_fiscal_year = row["record_fiscal_year"]
            record_fiscal_quarter = row["record_fiscal_quarter"]
            record_calendar_year = row["record_calendar_year"]
            record_calendar_quarter = row["record_calendar_quarter"]
            record_calendar_month = row["record_calendar_month"]
            record_calendar_day = row["record_calendar_day"]

            embed = disnake.Embed(title=f"🖋️ Accounting Service - Weekly Treasury Offerings", 
                                description='```py\nDetails of offerings for regular weekly Treasury Bills. All figures are rounded to the nearest million.```', 
                                color=disnake.Colour.dark_teal())

            embed.add_field(name=f"Record:", value=f"> Name: **Regular Weekly Treasury Bills**\n> Date: **{record_date}**\n> Issue Date: **{issue_date}**\n> Maturity Date: **{maturity_date}**")
            embed.add_field(name=f"Fiscal Info:", value=f"> Year: **{record_fiscal_year}**\n> Qtr: **{record_fiscal_quarter}**")
            embed.add_field(name=f"Calendar Info:", value=f"> Year: **{record_calendar_year}**\n> Qtr: **{record_calendar_quarter}**\n> Month: **{record_calendar_month}**\n> Day: **{record_calendar_day}**")
            embed.add_field(name=f"Bids Info:", value=f"> Tendered: **{format_large_number(bids_tendered)}**\n> Accepted Total: **{format_large_number(bids_acc_total)}**\n> Accepted on Competitive Basis: **{format_large_number(bids_acc_comp_basis)}**\n> Accepted on Non-Competitive Basis: **{format_large_number(bids_acc_noncomp_basis)}**")
            embed.add_field(name=f"Price and Rate Info:", value=f"> High Price Per Hundred: **{high_price_per_hundred}**\n> High Discount Rate: **{high_discount_rate}**\n> High Investment Rate: **{high_investment_rate}**")
            embed.add_field(name=f"Days to Maturity:", value=f"> **{days_to_maturity}**")
            
            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji="📥")
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('data/weekly_treasury_offerings.csv'), ephemeral=True)


        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(AccountingSelect()))

    @accounting.sub_command()
    async def marketable_securities_offerings(self, inter: disnake.AppCmdInter):
        """🖋️ Offerings of Marketable Securities Other Than Regular Weekly Treasury Bills"""

        await inter.response.defer()
        

        r = requests.get("https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/tb/pdo2_offerings_marketable_securities_other_regular_weekly_treasury_bills?sort=-record_date&limit=1000").json()
        data = r['data']

        df = pd.DataFrame(data)
        df.to_csv('data/marketable_securities_offerings.csv')
        embeds = []

        for i, row in df.iterrows():
            record_date = row["record_date"]
            auction_date = row["auction_date"]
            issue_date = row["issue_date"]
            securities_desc = row["securities_desc"]
            period_to_final_maturity = row["period_to_final_maturity"]
            tendered_mil_amt = float(row["tendered_mil_amt"]) * 1000000
            acc_mil_amt = float(row["acc_mil_amt"]) * 1000000
            acc_yield_discount_margin = row["acc_yield_discount_margin"]
            eq_price_for_notes_bonds = row["eq_price_for_notes_bonds"]
            src_line_nbr = row["src_line_nbr"]
            record_fiscal_year = row["record_fiscal_year"]
            record_fiscal_quarter = row["record_fiscal_quarter"]
            record_calendar_year = row["record_calendar_year"]
            record_calendar_quarter = row["record_calendar_quarter"]
            record_calendar_month = row["record_calendar_month"]
            record_calendar_day = row["record_calendar_day"]

            embed = disnake.Embed(title=f"🖋️ Accounting Service - Marketable Securities Offerings", 
                                description='```py\nDetails of offerings for marketable securities other than regular weekly Treasury Bills. All figures are rounded to the nearest million.```', 
                                color=disnake.Colour.dark_teal())

            embed.add_field(name=f"Record:", value=f"> Date: **{record_date}**\n> Description: **{securities_desc}**\n> Auction Date: **{auction_date}**\n> Issue Date: **{issue_date}**")
            embed.add_field(name=f"Fiscal Info:", value=f"> Year: **{record_fiscal_year}**\n> Qtr: **{record_fiscal_quarter}**")
            embed.add_field(name=f"Calendar Info:", value=f"> Year: **{record_calendar_year}**\n> Qtr: **{record_calendar_quarter}**\n> Month: **{record_calendar_month}**\n> Day: **{record_calendar_day}**")
            embed.add_field(name=f"Offerings Info:", value=f"> Tendered: **{format_large_number(tendered_mil_amt)}**\n> Accepted: **{format_large_number(acc_mil_amt)}**\n> Yield/Discount Margin: **{acc_yield_discount_margin}**\n> Eq. Price for Notes/Bonds: **{eq_price_for_notes_bonds}**")
            embed.add_field(name=f"Final Maturity:", value=f"> **{period_to_final_maturity}**")
            
            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji="📥")
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('data/marketable_securities_offerings.csv'), ephemeral=True)


        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(AccountingSelect()))


    @accounting.sub_command()
    async def federal_securities_distribution(self, inter: disnake.AppCmdInter):
        """🖋️ Distribution of Federal Securities by Class of Investors and Type of Issues"""

        await inter.response.defer()
        

        r = requests.get("https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/tb/ofs1_distribution_federal_securities_class_investors_type_issues?sort=-record_date&limit=1000").json()
        data = r['data']

        df = pd.DataFrame(data)
        df.to_csv('data/federal_securities_distribution.csv')
        embeds = []

        for i, row in df.iterrows():
            record_date = row["record_date"]
            end_fiscal_year_or_month = row["end_fiscal_year_or_month"]
            securities_classification = row["securities_classification"]
            investors_classification = row["investors_classification"]
            issues_type = row["issues_type"]
            securities_mil_amt = float(row["securities_mil_amt"]) * 1000000
            src_line_nbr = row["src_line_nbr"]
            record_fiscal_year = row["record_fiscal_year"]
            record_fiscal_quarter = row["record_fiscal_quarter"]
            record_calendar_year = row["record_calendar_year"]
            record_calendar_quarter = row["record_calendar_quarter"]
            record_calendar_month = row["record_calendar_month"]
            record_calendar_day = row["record_calendar_day"]

            embed = disnake.Embed(title=f"🖋️ Accounting Service - Federal Securities Distribution",
                                description='```py\nDistribution of Federal securities by class of investors and type of issues. All figures are rounded to the nearest million.```',
                                color=disnake.Colour.darker_grey())

            embed.add_field(name=f"Record:", value=f"> Date: **{record_date}**\n> Fiscal Year End: **{end_fiscal_year_or_month}**")
            embed.add_field(name=f"Fiscal Info:", value=f"> Year: **{record_fiscal_year}**\n> Qtr: **{record_fiscal_quarter}**")
            embed.add_field(name=f"Calendar Info:", value=f"> Year: **{record_calendar_year}**\n> Qtr: **{record_calendar_quarter}**\n> Month: **{record_calendar_month}**\n> Day: **{record_calendar_day}**")
            embed.add_field(name=f"Securities Info:", value=f"> Classification: **{securities_classification}**\n> Investors Classification: **{investors_classification}**\n> Issues Type: **{issues_type}**")
            embed.add_field(name=f"Amount:", value=f"> **{securities_mil_amt}**")

            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji="📥")
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('data/federal_securities_distribution.csv'), ephemeral=True)


        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(AccountingSelect()))



    @accounting.sub_command()
    async def estimated_ownership(self, inter: disnake.AppCmdInter):
        """🖋️ Estimated Ownership of Treasury Securities"""

        await inter.response.defer()
        

        r = requests.get("https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/tb/ofs2_estimated_ownership_treasury_securities?sort=-record_date&limit=1000").json()
        data = r['data']

        df = pd.DataFrame(data)
        df.to_csv('data/estimated_ownership.csv')
        embeds = []

        for i, row in df.iterrows():
            record_date = row["record_date"]
            end_of_month = row["end_of_month"]
            securities_owner = row["securities_owner"]
            securities_bil_amt = float(row["securities_bil_amt"]) * 1000000
            src_line_nbr = row["src_line_nbr"]
            record_fiscal_year = row["record_fiscal_year"]
            record_fiscal_quarter = row["record_fiscal_quarter"]
            record_calendar_year = row["record_calendar_year"]
            record_calendar_quarter = row["record_calendar_quarter"]
            record_calendar_month = row["record_calendar_month"]
            record_calendar_day = row["record_calendar_day"]

            embed = disnake.Embed(title=f"🖋️ Accounting Service - Estimated Ownership of Treasury Securities",
                                description='```py\nEstimated ownership of Treasury securities by various entities. All figures are rounded to the nearest billion.```',
                                color=disnake.Colour.darker_grey())

            embed.add_field(name=f"Record:", value=f"> Date: **{record_date}**\n> End of Month: **{end_of_month}**")
            embed.add_field(name=f"Fiscal Info:", value=f"> Year: **{record_fiscal_year}**\n> Qtr: **{record_fiscal_quarter}**")
            embed.add_field(name=f"Calendar Info:", value=f"> Year: **{record_calendar_year}**\n> Qtr: **{record_calendar_quarter}**\n> Month: **{record_calendar_month}**\n> Day: **{record_calendar_day}**")
            embed.add_field(name=f"Ownership Info:", value=f"> Owner: **{securities_owner}**")
            embed.add_field(name=f"Amount:", value=f"> **{securities_bil_amt}**")

            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji="📥")
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('data/estimated_ownership.csv'), ephemeral=True)

      
        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(AccountingSelect()))



    @accounting.sub_command()
    async def outstanding_circulation(self, inter: disnake.AppCmdInter):
        """🖋️ Amounts Outstanding in Circulation"""

        await inter.response.defer()
        

        r = requests.get("https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/tb/uscc1_amounts_outstanding_circulation?sort=-record_date&limit=1000").json()
        data = r['data']

        df = pd.DataFrame(data)
        df.to_csv('data/outstanding_circulation.csv')
        embeds = []

        for i, row in df.iterrows():
            # Converting all relevant values to float for numerical accuracy
            # Handle amounts outstanding
            for col in ['total_currency_coins_amt', 'total_currency_amt', 'federal_reserve_notes_amt', 
                        'us_notes_amt', 'currency_no_longer_issued_amt', 'total_coins_amt', 
                        'dollar_coins_amt', 'fractional_coins_amt']:
                value = row[col]
                if pd.isna(value):
                    row[col] = "N/A"  # Replace NaN with "N/A"
                else:
                    row[col] = float(value)  # Convert to float if it's not NaN


            embed = disnake.Embed(title=f"🖋️ Accounting Service - Amounts Outstanding in Circulation", 
                                description='```py\nData on the amounts of currency and coins outstanding in circulation.```', 
                                color=disnake.Colour.darker_grey())

            embed.add_field(name=f"Record:", value=f"> Date: **{row['record_date']}**\n> As of Date: **{row['currency_coins_as_of_date']}**")
            embed.add_field(name=f"Fiscal Info:", value=f"> Year: **{row['record_fiscal_year']}**\n> Qtr: **{row['record_fiscal_quarter']}**")
            embed.add_field(name=f"Calendar Info:", value=f"> Year: **{row['record_calendar_year']}**\n> Qtr: **{row['record_calendar_quarter']}**\n> Month: **{row['record_calendar_month']}**\n> Day: **{row['record_calendar_day']}**")
            embed.add_field(name=f"Amounts Outstanding:", value=f"> Total Currency and Coins: **{row['total_currency_coins_amt']}**\n> Total Currency: **{row['total_currency_amt']}**\n> Federal Reserve Notes: **{row['federal_reserve_notes_amt']}**")
            embed.add_field(name=f"More Details:", value=f"> US Notes: **{row['us_notes_amt']}**\n> No Longer Issued: **{row['currency_no_longer_issued_amt']}**\n> Total Coins: **{row['total_coins_amt']}**\n> Dollar Coins: **{row['dollar_coins_amt']}**\n> Fractional Coins: **{row['fractional_coins_amt']}**")

            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji="📥")
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('data/outstanding_circulation.csv'), ephemeral=True)

       
        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(AccountingSelect()))



    @accounting.sub_command()
    async def esf_balances(self, inter: disnake.AppCmdInter):
        """🖋️ Balances in the Exchange Stabilization Fund"""

        await inter.response.defer()
        

        r = requests.get("https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v1/accounting/tb/esf1_balances?sort=-record_date&limit=1000").json()
        data = r['data']

        df = pd.DataFrame(data)
        df.to_csv('data/esf_balances.csv')
        embeds = []

        for i, row in df.iterrows():
            # Handling null values
            activity_thous_amt = float(row["activity_thous_amt"]) if row["activity_thous_amt"] != "null" else "N/A"
            balance_thous_amt = float(row["balance_thous_amt"]) if row["balance_thous_amt"] != "null" else "N/A"

            embed = disnake.Embed(title=f"🖋️ Accounting Service - ESF Balances", 
                                description='```py\nBalances in the Exchange Stabilization Fund.```', 
                                color=disnake.Colour.darker_grey())

            embed.add_field(name=f"Record:", value=f"> Date: **{row['record_date']}**\n> Report Date: **{row['report_date']}**\n> Classification: **{row['classification_desc']}**")
            embed.add_field(name=f"Fiscal Info:", value=f"> Year: **{row['record_fiscal_year']}**\n> Qtr: **{row['record_fiscal_quarter']}**")
            embed.add_field(name=f"Calendar Info:", value=f"> Year: **{row['record_calendar_year']}**\n> Qtr: **{row['record_calendar_quarter']}**\n> Month: **{row['record_calendar_month']}**\n> Day: **{row['record_calendar_day']}**")
            embed.add_field(name=f"Amounts:", value=f"> Activity (Thousand): **{activity_thous_amt}**\n> Balance (Thousand): **{balance_thous_amt}**")

            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji="📥")
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('data/esf_balances.csv'), ephemeral=True)

     
        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(AccountingSelect()))

accounting = Accounting(commands.Bot)

class AccountingSelect(disnake.ui.Select):
    def __init__(self):
        super().__init__( 
            placeholder='Select A Function -->',
            min_values=1,
            max_values=1,
            custom_id='accountingSelect',
            options=[
                disnake.SelectOption(label='Weekly Treasuries', value="0", description='View weekly treasury offerings.', emoji="🖋️"),
                disnake.SelectOption(label='Marketable Securities', value="1", description='View marketable securities offerings.', emoji="🖋️"),
                disnake.SelectOption(label='Treasury Ownership', value="2", description='View estimated ownership of treasury securities.', emoji="🖋️"),
                disnake.SelectOption(label='Outstanding Circulation', value="3", description='View amounts outstanding in circulation.', emoji="🖋️"),
                disnake.SelectOption(label='ESF Balances', value="5", description='View balances in the Exchange Stabilization Fund.', emoji="🖋️"),
                disnake.SelectOption(label='Federal Securities Dist.', value="6", description=f"View federal securities distributions.")
                # Add more options here as needed
            ]
        )
        



    async def callback(self, inter: disnake.AppCmdInter):
        if self.values[0] == "0":
            await accounting.weekly_treasury_offerings(inter)
        elif self.values[0] == "1":
            await accounting.marketable_securities_offerings(inter)
        elif self.values[0] == "2":
            await accounting.estimated_ownership(inter)
        elif self.values[0] == "3":
            await accounting.outstanding_circulation(inter)
        elif self.values[0] == "5":
            await accounting.esf_balances(inter)
        elif self.values[0] == "6":
            await accounting.federal_securities_distribution(inter)

def setup(bot: commands.Bot):
    bot.add_cog(Accounting(bot))
    print('Accounting commands - READY!')



import aiohttp
import pandas as pd
class CapitalFlow:
    """
    A class representing capital flow data for a stock.

    Attributes:
        superin (float): The amount of super large inflow formatted with commas.
        superout (float): The amount of super large outflow formatted with commas.
        supernet (float): The amount of super large net flow formatted with commas.
        largein (float): The amount of large inflow formatted with commas.
        largeout (float): The amount of large outflow formatted with commas.
        largenet (float): The amount of large net flow formatted with commas.
        newlargein (float): The amount of new large inflow formatted with commas.
        newlargeout (float): The amount of new large outflow formatted with commas.
        newlargenet (float): The amount of new large net flow formatted with commas.
        newlargeinratio (float): The new large inflow ratio formatted as a percentage with 2 decimal places.
        newlargeoutratio (float): The new large outflow ratio formatted as a percentage with 2 decimal places.
        mediumin (float): The amount of medium inflow formatted with commas.
        mediumout (float): The amount of medium outflow formatted with commas.
        mediumnet (float): The amount of medium net flow formatted with commas.
        mediuminratio (float): The medium inflow ratio formatted as a percentage with 2 decimal places.
        mediumoutratio (float): The medium outflow ratio formatted as a percentage with 2 decimal places.
        smallin (float): The amount of small inflow formatted with commas.
        smallout (float): The amount of small outflow formatted with commas.
        smallnet (float): The amount of small net flow formatted with commas.
        smallinratio (float): The small inflow ratio formatted as a percentage with 2 decimal places.
        smalloutratio (float): The small outflow ratio formatted as a percentage with 2 decimal places.
        majorin (float): The amount of major inflow formatted with commas.
        majorinratio (float): The major inflow ratio formatted as a percentage with 2 decimal places.
        majorout (float): The amount of major outflow formatted with commas.
        majoroutratio (float): The major outflow ratio formatted as a percentage with 2 decimal places.
        majornet (float): The amount of major net flow formatted with commas.
        retailin (float): The amount of retail inflow formatted with commas.
        retailinratio (float): The retail inflow ratio formatted as a percentage with 2 decimal places.
        retailout (float): The amount of retail outflow formatted with commas.
        retailoutratio (float): The retail outflow ratio formatted as a percentage with 2 decimal places.

    Methods:
        async def capital_flow(id: str) -> CapitalFlow:
            Returns an instance of the CapitalFlow class for a given stock ticker ID.
            The data is fetched asynchronously using aiohttp.
    """

    def __init__(self, item):
        self.superin = float(item['superLargeInFlow']) if 'superLargeInFlow' in item else None
        self.superout = float(item['superLargeOutFlow']) if 'superLargeOutFlow' in item else None
        self.supernet = float(item['superLargeNetFlow']) if 'superLargeNetFlow' in item else None
        self.largein = float(item['largeInFlow']) if 'largeInFlow' in item else None
        self.largeout = float(item['largeOutFlow']) if 'largeOutFlow' in item else None
        self.largenet = float(item['largeNetFlow']) if 'largeNetFlow' in item else None
        self.newlargein = float(item['newLargeInFlow']) if 'newLargeInFlow' in item else None
        self.newlargeout = float(item['newLargeOutFlow']) if 'newLargeOutFlow' in item else None
        self.newlargenet = float(item['newLargeNetFlow']) if 'newLargeNetFlow' in item else None
        self.newlargeinratio = float(item['newLargeInFlowRatio']) if 'newLargeInFlowRatio' in item else None
        self.newlargeoutratio = float(item['newLargeOutFlowRatio']) if 'newLargeOutFlowRatio' in item else None
        self.mediumin = float(item['mediumInFlow']) if 'mediumInFlow' in item else None
        self.mediumout = float(item['mediumOutFlow']) if 'mediumOutFlow' in item else None
        self.mediumnet = float(item['mediumNetFlow']) if 'mediumNetFlow' in item else None
        self.mediuminratio = float(item['mediumInFlowRatio']) if 'mediumInFlowRatio' in item else None
        self.mediumoutratio = float(item['mediumOutFlowRatio']) if 'mediumOutFlowRatio' in item else None
        self.smallin = float(item['smallInFlow']) if 'smallInFlow' in item else None
        self.smallout = float(item['smallOutFlow']) if 'smallOutFlow' in item else None
        self.smallnet = float(item['smallNetFlow']) if 'smallNetFlow' in item else None
        self.smallinratio = float(item['smallInFlowRatio']) if 'smallInFlowRatio' in item else None
        self.smalloutratio = float(item['smallOutFlowRatio']) if 'smallOutFlowRatio' in item else None
        self.majorin = float(item['majorInFlow']) if 'majorInFlow' in item else None
        self.majorinratio = float(item['majorInFlowRatio']) if 'majorInFlowRatio' in item else None
        self.majorout = float(item['majorOutFlow']) if 'majorOutFlow' in item else None
        self.majoroutratio = float(item['majorOutFlowRatio']) if 'majorOutFlowRatio' in item else None
        self.majornet = float(item['majorNetFlow']) if 'majorNetFlow' in item else None
        self.retailin = float(item['retailInFlow']) if 'retailInFlow' in item else None
        self.retailinratio = float(item['retailInFlowRatio']) if 'retailInFlowRatio' in item else None
        self.retailout = float(item['retailOutFlow']) if 'retailOutFlow' in item else None
        self.retailoutratio = float(item['retailOutFlowRatio']) if 'retailOutFlowRatio' in item else None


        self.data_dict = {
            'superin': float(item['superLargeInFlow']) if 'superLargeInFlow' in item else None,
            'superout': float(item['superLargeOutFlow']) if 'superLargeOutFlow' in item else None,
            'supernet': float(item['superLargeNetFlow']) if 'superLargeNetFlow' in item else None,
            'largein': float(item['largeInFlow']) if 'largeInFlow' in item else None,
            'largeout': float(item['largeOutFlow']) if 'largeOutFlow' in item else None,
            'largenet': float(item['largeNetFlow']) if 'largeNetFlow' in item else None,
            'newlargein': float(item['newLargeInFlow']) if 'newLargeInFlow' in item else None,
            'newlargeout': float(item['newLargeOutFlow']) if 'newLargeOutFlow' in item else None,
            'newlargenet': float(item['newLargeNetFlow']) if 'newLargeNetFlow' in item else None,
            'newlargeinratio': float(item['newLargeInFlowRatio']) if 'newLargeInFlowRatio' in item else None,
            'newlargeoutratio': float(item['newLargeOutFlowRatio']) if 'newLargeOutFlowRatio' in item else None,
            'mediumin': float(item['mediumInFlow']) if 'mediumInFlow' in item else None,
            'mediumout': float(item['mediumOutFlow']) if 'mediumOutFlow' in item else None,
            'mediumnet': float(item['mediumNetFlow']) if 'mediumNetFlow' in item else None,
            'mediuminratio': float(item['mediumInFlowRatio']) if 'mediumInFlowRatio' in item else None,
            'mediumoutratio': float(item['mediumOutFlowRatio']) if 'mediumOutFlowRatio' in item else None,
            'smallin': float(item['smallInFlow']) if 'smallInFlow' in item else None,
            'smallout': float(item['smallOutFlow']) if 'smallOutFlow' in item else None,
            'smallnet': float(item['smallNetFlow']) if 'smallNetFlow' in item else None,
            'smallinratio': float(item['smallInFlowRatio']) if 'smallInFlowRatio' in item else None,
            'smalloutratio': float(item['smallOutFlowRatio']) if 'smallOutFlowRatio' in item else None,
            'majorin': float(item['majorInFlow']) if 'majorInFlow' in item else None,
            'majorinratio': float(item['majorInFlowRatio']) if 'majorInFlowRatio' in item else None,
            'majorout': float(item['majorOutFlow']) if 'majorOutFlow' in item else None,
            'majoroutratio': float(item['majorOutFlowRatio']) if 'majorOutFlowRatio' in item else None,
            'majornet': float(item['majorNetFlow']) if 'majorNetFlow' in item else None,
            'retailin': float(item['retailInFlow']) if 'retailInFlow' in item else None,
            'retailinratio': float(item['retailInFlowRatio']) if 'retailInFlowRatio' in item else None,
            'retailout': float(item['retailOutFlow']) if 'retailOutFlow' in item else None,
            'retailoutratio': float(item['retailOutFlowRatio']) if 'retailOutFlowRatio' in item else None
        }

        self.df = pd.DataFrame(self.data_dict, index=[0]).transpose()

    async def fetch_data(self, id):
        async with aiohttp.ClientSession() as session:
            async with session.get(f"https://quotes-gw.webullfintech.com/api/stock/capitalflow/ticker?tickerId={id}&showHis=true") as response:
                r = await response.json()
                latest = r.get('latest', None)
                if latest is not None:
                    item = latest.get('item', None)
                    if item is not None:
                        self.superin = "{:,}".format(item.get('superLargeInflow', 0))
                        self.superout = "{:,}".format(item.get('superLargeOutflow', 0))
                        self.supernet = "{:,}".format(item.get('superLargeNetFlow', 0))
                        self.largein = "{:,}".format(item.get('largeInflow', 0))
                        self.largeout = "{:,}".format(item.get('largeOutflow', 0))
                        self.largenet = "{:,}".format(item.get('largeNetFlow', 0))
                        self.newlargein = "{:,}".format(item.get('newLargeInflow', 0))
                        self.newlargeout = "{:,}".format(item.get('newLargeOutflow', 0))
                        self.newlargenet = "{:,}".format(item.get('newLargeNetFlow', 0))
                        self.newlargeinratio = "{:,.2f}".format(float(item.get('newLargeInflowRatio', 0)) * 100)
                        self.newlargeoutratio = "{:,.2f}".format(float(item.get('newLargeOutflowRatio', 0)) * 100)
                        self.mediumin = "{:,}".format(item.get('mediumInflow', 0))
                        self.mediumout = "{:,}".format(item.get('mediumOutflow', 0))
                        self.mediumnet = "{:,}".format(item.get('mediumNetFlow', 0))
                        self.mediuminratio = "{:,.2f}".format(float(item.get('mediumInflowRatio', 0)) * 100)
                        self.mediumoutratio = "{:,.2f}".format(float(item.get('mediumOutflowRatio', 0)) * 100)
                        self.smallin = "{:,}".format(item.get('smallInflow', 0))
                        self.smallout = "{:,}".format(item.get('smallOutflow', 0))
                        self.smallnet = "{:,}".format(item.get('smallNetFlow', 0))
                        self.smallinratio = "{:,.2f}".format(float(item.get('smallInflowRatio', 0)) * 100)
                        self.smalloutratio = "{:,.2f}".format(float(item.get('smallOutflowRatio', 0)) * 100)
                        self.majorin = "{:,}".format(item.get('majorInflow', 0))
                        self.majorinratio = "{:,.2f}".format(float(item.get('majorInflowRatio', 0)) * 100)
                        self.majorout = "{:,}".format(item.get('majorOutflow', 0))
                        self.majoroutratio = "{:,.2f}".format(float(item.get('majorOutflowRatio', 0)) * 100)
                        self.majornet = "{:,}".format(item.get('majorNetFlow', 0))
                        self.retailin = "{:,}".format(item.get('retailInflow', 0))
                        self.retailinratio = "{:,.2f}".format(float(item.get('retailInflowRatio', 0)) * 100)
                        self.retailout = "{:,}".format(item.get('retailOutflow', 0))
                        self.retailoutratio = "{:,.2f}".format(float(item.get('retailOutflowRatio', 0)) * 100)


from dataclasses import dataclass, asdict
from datetime import datetime

import pandas as pd

@dataclass
class InsiderTrades:
    ticker: str
    name: str
    relationship: str
    date: datetime
    transaction: str
    cost: float
    shares: int
    value_dollar: float
    shares_total: int
    date_filled: datetime
    empty_string: str  # This field may need special handling


    def as_dataframe(self) -> pd.DataFrame:
        """Converts the object to a pandas DataFrame."""
        return pd.DataFrame([asdict(self)])

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO insider_transaction (ticker, name, relationship, date, transaction, cost, shares, value_dollar, shares_total, date_filled, empty_string) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.ticker, self.name, self.relationship, self.date, self.transaction, self.cost, self.shares, self.value_dollar, self.shares_total, self.date_filled, self.empty_string))
        conn.commit()



from dataclasses import dataclass,asdict

import pandas as pd

@dataclass
class LowFloatData:
    rank: int
    ticker: str
    company_name: str
    exchange: str
    previous_close: float
    one_day_change: float
    floating_shares: int
    outstanding_shares: int
    short_int: float
    market_cap: int
    industry: str

    def as_dataframe(self) -> pd.DataFrame:
        """Converts the object to a pandas DataFrame."""
        return pd.DataFrame([asdict(self)])

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO low_float (rank, ticker, company_name, exchange, previous_close, one_day_change, floating_shares, outstanding_shares, short_int, market_cap, industry) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.rank, self.ticker, self.company_name, self.exchange, self.previous_close, self.one_day_change, self.floating_shares, self.outstanding_shares, self.short_int, self.market_cap, self.industry))
        conn.commit()



import sys
from pathlib import Path

# Add the project fetch to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)

import pandas as pd
from .webull_helpers import flatten_list_of_dicts, flatten_dict


class WebullOptionsData:
    def __init__(self, data):
        self.as_dataframe = None
      # Initialize the remaining attributes
        self.expireDate = [i.get('expireDate') for i in data] if isinstance(data, list) else data.get('expireDate')
        self.tickerId = [i.get('tickerId') for i in data] if isinstance(data, list) else data.get('tickerId')
        self.belongTickerId = [i.get('belongTickerId') for i in data] if isinstance(data, list) else data.get('belongTickerId')
        self.openIntChange = [i.get('openIntChange') for i in data] if isinstance(data, list) else data.get('openIntChange')
        self.activeLevel = [i.get('activeLevel') for i in data] if isinstance(data, list) else data.get('activeLevel')
        self.direction = [i.get('direction') for i in data] if isinstance(data, list) else data.get('direction')
        self.symbol = [i.get('symbol') for i in data] if isinstance(data, list) else data.get('symbol')
        self.unSymbol = [i.get('unSymbol') for i in data] if isinstance(data, list) else data.get('unSymbol')
        self.askList = [i.get('askList', None) for i in data] if isinstance(data, list) else data.get('askList', None)
        self.bidList = [i.get('bidList', None) for i in data] if isinstance(data, list) else data.get('bidList', None)
        self.strikePrice = [float(i.get('strikePrice', None)) for i in data] if isinstance(data, list) else float(data.get('strikePrice', None))
        self.open = [i.get('open') for i in data]
        self.high = [i.get('high') for i in data]
        self.low = [i.get('low') for i in data]
        self.close =  [i.get('close') for i in data]
        self.preClose = [float(item.get('preClose')) if isinstance(item.get('preClose'), (int, float)) else None for item in data] if isinstance(data, list) else float(data.get('preClose')) if isinstance(data.get('preClose'), (int, float)) else None

        self.change = [float(item.get('change')) if isinstance(item.get('change'), (int, float)) else None for item in data] if isinstance(data, list) else float(data.get('change')) if isinstance(data.get('change'), (int, float)) else None

        self.changeRatio = [round(float(item.get('changeRatio')) * 100, 2) if isinstance(item.get('changeRatio'), (int, float)) else None for item in data] if isinstance(data, list) else round(float(data.get('changeRatio')) * 100, 2) if isinstance(data.get('changeRatio'), (int, float)) else None

        self.volume = [i.get('volume') for i in data]


        self.latestPriceVol = [float(item.get('latestPriceVol')) if isinstance(item.get('latestPriceVol'), (int, float)) else None for item in data] if isinstance(data, list) else float(data.get('latestPriceVol')) if isinstance(data.get('latestPriceVol'), (int, float)) else None

        self.openInterest = [float(item.get('openInterest')) if isinstance(item.get('openInterest'), (int, float)) else None for item in data] if isinstance(data, list) else float(data.get('openInterest')) if isinstance(data.get('openInterest'), (int, float)) else None

        self.openIntChange = [float(item.get('openIntChange')) if isinstance(item.get('openIntChange'), (int, float)) else None for item in data] if isinstance(data, list) else float(data.get('openIntChange')) if isinstance(data.get('openIntChange'), (int, float)) else None

                # For the following, it seems they are expected to be strings, so we'll leave them as is, just checking for lists.
        self.impVol = [float(item.get('impVol')) if item.get('impVol', '--') != '--' else None for item in data] if isinstance(data, list) else (float(data.get('impVol')) if data.get('impVol', '--') != '--' else None)

        self.delta = [float(item.get('delta')) if item.get('delta', '--') != '--' else None for item in data] if isinstance(data, list) else (float(data.get('delta')) if data.get('delta', '--') != '--' else None)

        self.gamma = [float(item.get('gamma')) if item.get('gamma', '--') != '--' else None for item in data] if isinstance(data, list) else (float(data.get('gamma')) if data.get('gamma', '--') != '--' else None)

        self.theta = [float(item.get('theta')) if item.get('theta', '--') != '--' else None for item in data] if isinstance(data, list) else (float(data.get('theta')) if data.get('theta', '--') != '--' else None)

        self.rho = [float(item.get('rho')) if item.get('rho', '--') != '--' else None for item in data] if isinstance(data, list) else (float(data.get('rho')) if data.get('rho', '--') != '--' else None)

        self.vega = [float(item.get('vega')) if item.get('vega', '--') != '--' else None for item in data] if isinstance(data, list) else (float(data.get('vega')) if data.get('vega', '--') != '--' else None)


        self.askList = [i.get('askList', None) for i in data] if isinstance(data, list) else data.get('askList', None)
        self.bidList = [i.get('bidList', None) for i in data] if isinstance(data, list) else data.get('bidList', None)
        self.bidPrice = [i[0].get('price') for i in self.bidList if i and i[0] is not None]

  
        self.askPrice = [i[0].get('price') for i in self.askList if i and i[0] is not None]
        self.bidVolume = [i[0].get('volume') for i in self.bidList if i and i[0] is not None]
        self.askVolume = [i[0].get('volume') for i in self.askList if i and i[0] is not None]
        self.bidExchange = [i[0].get('quoteEx') for i in self.bidList if i and i[0] is not None]
        self.askExchange = [i[0].get('quoteEx') for i in self.askList if i and i[0] is not None]

     
        self.data_dict = { 
            'strike_price': self.strikePrice,
            'call_put': self.direction,
            'expiry_date': self.expireDate,
            'option_symbol': self.symbol,
            'option_id': self.tickerId,
            'ticker_id': self.belongTickerId,  # Assuming you meant belongTickerId here for the underlying ticker
            'open': self.open,
            'high': self.high,
            'low': self.low,
            'close': self.close,
            'pre_close': self.preClose,
            'change': self.change,
            'change_ratio': self.changeRatio,
            'volume': self.volume,
            'bid_price': self.bidPrice,
            'bid_volume': self.bidVolume,
            'bid_exchange': self.bidExchange,
            'ask_price': self.askPrice,
            'ask_volume': self.askVolume,
            'ask_exchange': self.askExchange,
            'latest_price_vol': self.latestPriceVol,
            'open_interest': self.openInterest,
            'open_interest_change': self.openIntChange,
            'implied_volatility': self.impVol,
            'delta': self.delta,
            'vega': self.vega,
            'gamma': self.gamma,
            'theta': self.theta,
            'rho': self.rho,

            # ... and any other attributes you wish to include
        }

        try:
            self.as_dataframe = pd.DataFrame(self.data_dict)

        except ValueError:
            print(f'Arrays not same length - skipping!')



class VolumeAnalysis:
    def __init__(self, data):
        self.ticker_id = data.get('belongTickerId', None)
        self.option_id = data.get('tickerId', None)
        self.total_trades = data.get('totalNum', None)
        self.total_volume = data.get('totalVolume', None)
        self.avg_price = data.get('avgPrice', None)
        self.buy_volume = data.get('buyVolume', None)
        self.sell_volume = data.get('sellVolume', None)
        self.neutral_volume = data.get('neutralVolume', None)

        
        trades = flatten_list_of_dicts(data.get('datas', None))
        dates = flatten_list_of_dicts(data.get('dates', None))


        if trades is not None and dates is not None:
            flattened_data = []
            for trade, date in zip(trades, dates):
                # Flatten the date dictionary
                date_value = date[''] if '' in date else None
                
                # Combine the trade data with the date
                self.flattened_trade = {
                    'date': date_value,
                    **trade  # Unpack the trade dictionary
                }
                
                # Add the flattened trade to the list
                flattened_data.append(self.flattened_trade)

                # Create a DataFrame
                df = pd.DataFrame(flattened_data)

                df['option_id'] = self.option_id
                df.set_index('option_id', inplace=True)

            print(flattened_data)
        self.trades_and_dates_dict = { 

            'trades': trades,
            'dates': dates
        }

        self.data_dict = { 

            'ticker_id': self.ticker_id,
            'option_id': self.option_id,
            'total_trades': self.total_trades,
            'total_volume': self.total_volume,
            'avg_price': self.avg_price,
            'buy_volume': self.buy_volume,
            'sell_volume': self.sell_volume,
            'neutral_volume': self.neutral_volume,

        }


        self.as_dataframe = pd.DataFrame(self.data_dict, index=[0])

        print(self.as_dataframe)


import os
from dotenv import load_dotenv
load_dotenv()
import asyncio
from apis.webull.webull_trading import WebullTrading
from apis.webull.webull_helpers import calculate_countdown,calculate_setup
from discord_webhook import AsyncDiscordWebhook,DiscordEmbed
from apis.polygonio.async_polygon_sdk import Polygon
from list_sets.dicts import hex_color_dict
from list_sets.ticker_lists import most_active_tickers
from discord_webhook import AsyncDiscordWebhook
from aiohttp.client_exceptions import ClientConnectorError, ClientOSError
import httpx




min1_td9= os.environ.get('min1_td9')
min5_td9=os.environ.get('min5_td9')
min15_td9= os.environ.get('min15_td9')
min20_td9=os.environ.get('min20_td9')
min30_td9= os.environ.get('min30_td9')
min60_td9= os.environ.get('min60_td9')
min120_td9=os.environ.get('min120_td9')
min240_td9= os.environ.get('min240_td9')
day_td9=os.environ.get('day_td9')



polygon = Polygon()



async def scan_for_td9(dataframe):
    if len(dataframe) < 13:
        return False
    for i in range(4, 13):
        if not all(dataframe['Close'].iloc[i] < dataframe['Close'].iloc[i - 4] for i in range(4, 13)):
            return False
    return True
most_active_tickers = set(most_active_tickers)


trading = WebullTrading()




timeframes = ['m1','m5','m15','m20','m30','m60','m120','m240','d1']


async def scan_bars(ticker):
    while True:
        
        all_results=[]
        tasks = []
        for timeframe in timeframes:
            try:
                bars = await trading.get_bars(ticker, timeframe)
                print(type(bars))  # Add this line to print out the type of bars
                print(bars) 

                last_thirteen_candles = bars[:13]
                result = await scan_for_td9(last_thirteen_candles)

                if result == True:
                        # Determine the correct timespan for RSI based on the TD9 timespan
                    if timeframe == 'm1':
                        rsi_timespan = 'minute'
                    elif timeframe in ['m5', 'm10', 'm15', 'm20', 'm30', 'm60']:
                        rsi_timespan = 'hour'
                    else:
                        rsi_timespan = 'day'  # or any default you deem appropriate

                    # Fetch the RSI with the determined timespan
                    rsi = await polygon.rsi(ticker, timespan=rsi_timespan, limit=1)

                    rsi = rsi.rsi_value[0]

                    if rsi <= 30:
                        color = hex_color_dict['green']

                    elif rsi >= 70:
                        color = hex_color_dict['red']

                    else:
                        color = hex_color_dict['grey']

                    embed = DiscordEmbed(title=f"TD9 | {ticker} | {timeframe}", description=f"```py\nTD9 is a technical analysis indicator that signals potential trend reversal after 9 consecutive price bars close higher or lower than the close 4 bars earlier. Traders capitalize on it by anticipating reversals, buying after bearish TD9, and selling after bullish TD9 setups. These embeds are color-coded to signify oversold or overbought RSI + TD9.```", color=color)
                    embed.add_embed_field(name=f"RSI: {rsi_timespan}", value=f"> **{rsi}**")
                    if result == True and timeframe == 'm1':
                        hook = AsyncDiscordWebhook(min1_td9, content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()

                    if result == True and timeframe == 'm5':
                        hook = AsyncDiscordWebhook(min5_td9, content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()


                    # elif result == True and timeframe == 'm10':
                    #     hook = AsyncDiscordWebhook(min, content=f"> # TD9 - **{ticker} | {timeframe}**")
                    #     await hook.execute()

                    

                    if result == True and timeframe == 'm15':
                        hook = AsyncDiscordWebhook(min15_td9,content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()


                    elif result == True and timeframe == 'm20':
                        hook = AsyncDiscordWebhook(min20_td9,content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()


                    if result == True and timeframe == 'm30':
                        hook = AsyncDiscordWebhook(min30_td9, content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()


                    if result == True and timeframe == 'm60':
                        hook = AsyncDiscordWebhook(min60_td9, content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()


                    if result == True and timeframe == 'm120':
                        hook = AsyncDiscordWebhook(min120_td9, content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()

                    if result == True and timeframe == 'm240':
                        hook = AsyncDiscordWebhook(min240_td9, content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()


                    if result == True and timeframe == 'd1':
                        hook = AsyncDiscordWebhook(day_td9, content=f"<@375862240601047070>")
                        hook.add_embed(embed)
                        await hook.execute()
            except (TypeError, AttributeError, KeyError, ClientConnectorError,ClientOSError, httpx.ConnectError, httpx.ReadError):
                continue

        await asyncio.gather(*tasks)
                
        

async def scan_all_bars():
    tickers = ['AAPL', 'MSFT', 'SPX', 'SPY', 'U', 'W', 'NVDA', 'AMD', 'SPXU', 'QQQ', 'TSLA', 'AVGO', 'RBLX', 'TQQQ', 'SOXS', 'SPXL', 'MSFT', 'GOOG', 'GOOGL', 'TAL', 'EDU', 'GOTU', 'TUYA']


    tasks = [scan_bars(i) for i in most_active_tickers]


    return await asyncio.gather(*tasks)




asyncio.run(scan_all_bars())


import sys
from pathlib import Path
# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)
import os
import sys
from pathlib import Path
import re
import disnake
from disnake.ext.commands.errors import CommandInvokeError
from pytrends.request import TrendReq
from disnake.ext import commands
from tabulate import tabulate
from asyncpg.exceptions import UniqueViolationError
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from apis.webull.webull_screener import WebulScreener, ScreenerSelect
from datetime import datetime
from apis.openai_ import OpenAISDK
from apis.polygonio.polygon_options import PolygonOptions
import disnake
from apis.y_finance.yf_sdk import yfSDK
import base64
from disnake.ext import commands
from _markets.list_sets.ticker_lists import most_active_tickers
most_active_tickers = set(most_active_tickers)
from discord_.bot_menus.modals.options_modal import OptionsDataModal
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
import asyncio
from discord_.bot_menus.pagination import AlertMenus
import aiohttp
from openai import OpenAI
import matplotlib.pyplot as plt
import io
import asyncpg
import pandas as pd

from apis.polygonio.polygon_options import PolygonOptions
from apis.webull.opt_modal import OptionModal, SQLQueryModal
from apis.y_finance.yf_sdk import yfSDK
from apis.gexbot.gexbot import GEXBot
SEC_FILINGS=[1153827348454584443, 1153827546706747455, 1153828102389104771, 1153828288372949062, 1153828427342807110, 1153828752183283752, 1153828753756135424, 1153829229943865344, 1153829615874355240, 1153833634097278976, 1156987190400786452, 1157034088310509568, 1157038070156234843, 1157038883184320624, 1157107615994761236, 1157107859688013866, 1174774750741020752, 1175108609751916606, 1175109001390866473, 1175109208987947108, 1175109603126693958, 1175110342884462604, 1175111342861058099, 1175112249061421207, 1175112431228424262, 1175112609222119467, 1175113168993923082, 1175113782083727504, 1175115817722073168, 1178053558239756288]


MOMENTUM_SCALPS =  [1136061999751631000, 1113439351246946415, 1113563471141941420, 1153564391606849568]


TRADYTICS =  [1112740350059098162, 1112740518753996851, 1112740590845702305, 1112755591358722199, 1112755644357947523, 1112768180469891255, 1112768263911387157, 1151947923970605076, 1112768289303711945, 1112768500923109416, 1112768636961181717, 1151947255167856680, 1151947507069366363]

SECTOR_DARKPOOLS =  [1152761279593840640, 1152761281200279632, 1152761528035070033, 1153325170451288064, 1152761724752105513, 1152761278176170046, 1152761733920858152, 1152761735275630632, 1152761276913680524, 1152761736630390925, 1152761897033154641, 1152761738060644483]

RSI_FEEDS =  [1124185351477526539, 1131240086378393600, 1152294859303043152, 1131238736353562745, 1176615805941723259, 1162153100828754053, 1162153373693399121]

OPTION_FEEDS =  [1134556646278975599, 1154854338167066634, 1077621690131021894, 1165777024091160596, 1156649967214137404, 1165783348443086878, 1156645245287673988, 1156645246847963208, 1156645248932515910, 1156645254460608613, 1165783313907195994, 1165781494040641536, 1165781495739330620, 1165781498096521306, 1165781499212210236, 1167521435514847312]

REDDIT =  [1152766178385731681, 1152766191367098450, 1152766193464262736, 1152766195100024832, 1152766447089627146, 1152968556720431115, 1152973376017682482]

RSS_FEEDS = [1028667813168173167, 1019360302250332301, 1028668345702166698, 1053789985909768254, 1042559047896932362, 1019360339856470146, 1053789646926123086, 1053789818007584938, 1053789896352993341]

STOCK_CONDITIONS = [1149131544422776883, 1149131587670257776, 1161659695157756074, 1161661864443396186, 1161662498278228019, 1161663210236162149, 1161663337726230628, 1174718171459108935]

OPTIONS_CONDITIONS = [1113143417334157332, 1113149881352192030, 1161370852055601163, 1113144168567537795, 1113146480455331840, 1129489929773260872, 1113147180140728383, 1129885678063337474, 1133412240331112448, 1133412527406059600, 1133426263546134568, 1133417470363971794, 1133451695985266738, 1134554830627672135, 1129905458480689242, 1148975826382102568, 1151231796994904094, 1155882061081612318, 1155894099770085476, 1155894448992034857, 1155894842736529490, 1155895232727089162, 1174013415073775726]

OPEN_BB = [1089364481672478780, 1035273203683172434, 1089364148535701544, 1089364318274990191]

STOCK_FEEDS = [1149077047331790848, 1149077345458737172, 1148283428471582870, 1149074180013310093, 1151224045212270732, 1151223877108781169, 1149081836727828550, 1149081838569148497, 1148283426730934303, 1152279428219555840, 1152279430115381248, 1153468681901322460, 1153468687311970384, 1170147962396102757]

CRYPTO = [1162159268598911047, 1178482490864832594]
td9_ids = [
    int("1158471263984029777"),  # Channel: td9⏺5minute
    int("1158488492163211264"),  # Channel: td9⏺day
    int("1158867480853352583"),  # Channel: td9⏺15minute
    int("1158867482824675358"),  # Channel: td9⏺30minute
    int("1158868591911899298"),  # Channel: td9⏺hour
    int("1158868593967108096"),  # Channel: td9⏺2hr
    int("1158868595619672135"),  # Channel: td9⏺4hr
    int("1161334711197630566"),  # Channel: td9⏺20minute
    int("1151905252392575067"),  # Channel: td9[]Minute
    int("1178480622596018226"), # td9 week
    int("1178480910623047822"), # td9 month
]
# List of channel IDs as integers
opt_vol_ids = [
    int("1156645245287673988"),  # Channel: 500➖1k➖vol
    int("1156645246847963208"),  # Channel: 1k➖10k➖vol
    int("1156645248932515910"),  # Channel: 10k➖50k➖vol
    int("1156645254460608613"),
    int("1154854338167066634")   # Channel: 50k➕vol
]

db_config = {
    "host": os.environ.get('DB_HOST', 'localhost'), # Default to this IP if 'DB_HOST' not found in environment variables
    "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
    "user": os.environ.get('DB_USER'), # Default to 'postgres' if 'DB_USER' not found
    "password": os.environ.get('DB_PASSWORD', 'fud'), # Use the password from environment variable or default
    "database": os.environ.get('DB_NAME', 'polygon') # Database name for the new jawless database
}
opts = PolygonOptions()
gexbot = GEXBot()
bot = commands.Bot(command_prefix="!", intents=disnake.Intents.all())
gptsdk=OpenAISDK()
yf = yfSDK()

from _markets.list_sets.ticker_lists import gex_tickers
from typing import List
import disnake
from disnake.ext import commands
import openai

import os
from dotenv import load_dotenv
load_dotenv()


# Function to format timestamp
def format_timestamp(ts):
    if isinstance(ts, str):
        return datetime.fromisoformat(ts).strftime('%y-%m-%d')
    elif isinstance(ts, datetime):
        return ts.strftime('%y-%m-%d')
    else:
        return ts  # In case it's neither a string nor a datetime object

# Apply formatting to td9_data
# Initialize the OpenAI client with your API key
openai.api_key = os.getenv("YOUR_OPENAI_KEY")

import openai



td9_timespan_dict = {
    'm1': '1 minute',
    'm5': '5 minute',
    'm15': '15 minute',
    'm30': '30 minute',
    'm60': '1 hour',
    'm120': '2 hour',
    'm240': '4 hour',
    'd1': 'daily',
    'w': 'weekly',
    'm': 'monthly'
}

# @bot.event
# async def on_ready():
#     guild_id = 888488311927242753  # Replace with your guild's ID
#     guild = bot.get_guild(guild_id)

#     if guild is None:
#         print("Guild not found")
#         return

#     channels = await guild.fetch_channels()

#     for channel in channels:
#         try:
#             # Attempt to process each channel
#             print(f"Channel: {channel.name} - ID: {channel.id}")
#         except disnake.errors.InvalidData as e:
#             print(f"Encountered an error with channel {channel.id}: {e}")
#             # Continue with the next channel
#             continue

# now = datetime.now()
# @bot.event
# async def on_message(message: disnake.Message):
#     """Use  GPT4 Vision to listen for image URLs"""

#     if message.author.id == 800519694754644029:
#         message = message.content.split('>')


#     if message.channel.id == 896207280117264434:
#         content = message.content
#         username = message.author.name
#         timestamp = now.strftime('%Y-%m-%d %H:%M:%S')
#         # Assuming you have a connection pool or a single connection to your database
#         conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
        
#         # Insert data into the market_data table
#         await conn.execute('''
#             INSERT INTO messages(message, username, timestamp)
#             VALUES($1, $2, $3)
#             ''', content, username, timestamp)
        
#         # Close the connection if you're not using a connection pool
#         await conn.close()

#     embeds = message.embeds
#     if embeds is not None and len(embeds) > 0:
#         title = embeds[0].title
#         description = embeds[0].description


#         if title is not None and 'Flow' in title:
    
#             ticker = title.split(": ")[1]
    

#             fields = embeds[0].fields

#             names = [i.name for i in fields]
#             values = [i.value for i in fields]

        
#             sentiment = values[1].split(' :')[0].lower()

#             data = values[3]

#             data  = data.split('\n')

#             data = [i.replace(",","").replace('$','').replace('>','').replace('%','').lower().split(': ')[1] for i in data]
  
#             if len(data) == 9:
            
#                 flow_type = title
#                 strike = data[0]
#                 call_put = data[1]
#                 expiry = data[2]
#                 side = data[3].split(' :')[0]
#                 volume = data[4]
#                 oi = data[5]
#                 multiplier=data[6]
#                 iv = data[7]
#                 dte = data[8]
            
#                 timestamp = now.strftime('%Y-%m-%d %H:%M:%S')
#                 try:
#                     # Assuming you have a connection pool or a single connection to your database
#                     conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
                    
#                     # Insert data into the market_data table
#                     await conn.execute('''
#                         INSERT INTO flow(flow_type, ticker, strike, call_put, expiry, side, volume, oi, multiplier, iv, dte, timestamp, sentiment)
#                         VALUES($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13)
#                         ''', flow_type, ticker, strike, call_put, expiry, side, volume, oi, multiplier, iv, dte, timestamp, sentiment)
                    

#                 except UniqueViolationError:
#                     pass


#         if title is not None and 'Bullseye' in title:
#             print(f"BULLSEYE: {title}")

#         if title is not None and 'RSI |' in title:
#             footer = embeds[0].footer.text
            


#             footer = footer.split(' | ')

#             ticker = footer[0]
#             rsi = footer[1].replace(']', '').replace('[', '')
#             timespan = footer[2]
#             status = footer[3]
#             timestamp = embeds[0].timestamp
#             timestamp = str(timestamp.astimezone()).split('.')[0]
#             # Convert string to datetime object
#             timestamp_dt = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
#             try:
#                 # Assuming you have a connection pool or a single connection to your database
#                 conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
                
#                 # Insert data into the market_data table
#                 await conn.execute('''
#                     INSERT INTO rsi_status(ticker, rsi, timespan, status, timestamp)
#                     VALUES($1, $2, $3, $4, $5)
#                     ''', ticker, rsi, timespan, status, timestamp_dt)
                
#                 # Close the connection if you're not using a connection pool
#                 await conn.close()
#             except UniqueViolationError:
#                 pass


            # if message.content.endswith('.jpg'):
        #     if message.author == bot.user:
        #         return
        #     url = message.content
        #     analysis = gptsdk.analyze_stock(url)

        #     embed = disnake.Embed(title=f"GPT Vision", description=f"> **{analysis}**", color=disnake.Colour.random())
        #     embed.set_thumbnail(url)

        #     await message.channel.send(embed=embed)


        # elif message.content.endswith('.png'):
        #     if message.author == bot.user:
        #         return
        #     url = message.content
        #     analysis = gptsdk.analyze_stock(url)

        #     embed = disnake.Embed(title=f"GPT Vision", description=f"> **{analysis}**", color=disnake.Colour.random())
        #     embed.set_thumbnail(url)

        #     await message.channel.send(embed=embed)




        # if message.attachments:
        #     print(message.content)
        #     url = message.attachments[0].proxy_url
        #     analysis = gptsdk.analyze_stock(url)

        #     embed = disnake.Embed(title=f"GPT Vision", description=f"> **{analysis}**", color=disnake.Colour.random())
        #     embed.set_thumbnail(url)

        #     await message.channel.send(embed=embed)


    #     if message.channel.id in REDDIT:
    #         embeds = message.embeds[0]
    #         title = embeds.title
    #         description = embeds.description
    #         channel = message.channel.name

    #         # Assuming you have a connection pool or a single connection to your database
    #         conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
            
    #         # Insert data into the market_data table
    #         await conn.execute('''
    #             INSERT INTO reddit_posts(title, context, subreddit, insertion_timestamp)
    #             VALUES($1, $2, $3, NOW())
    #             ''', title, description, channel)
            

    #     if message.channel.id in TRADYTICS:
    #         embeds = message.embeds[0]
    #         title = embeds.title
    #         desc = embeds.description
    #         fields = embeds.fields
    #         field_names = [i.name for i in fields]
    #         field_values = [i.value for i in fields]

            
    #         print(f"FIELD NAMES: {field_names}")
    #         print(f"FIELD VALUES: {field_values}")

    #     if message.channel.id in td9_ids:
    #         footer = message.embeds[0].footer.text

    #         footer = footer.split('|')

    #         symbol = footer[0].replace(' ', '')
    #         timespan = td9_timespan_dict.get(footer[1].replace(' ', ''))
    #         status = footer[2].replace(' ', '')

    #         # Assuming you have a connection pool or a single connection to your database
    #         conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
            
    #         # Insert data into the market_data table
    #         await conn.execute('''
    #             INSERT INTO market_data(ticker, timespan, td9_state, insertion_timestamp)
    #             VALUES($1, $2, $3, NOW())
    #             ''', symbol, timespan, status)

        
    #     elif message.channel.id in SEC_FILINGS:

    #         footer = message.embeds[0].footer.text
    #         footer = footer.split('|')
    #         ticker = footer[0].replace(' ','')
    #         title = footer[1]
    #         link = footer[2].replace(' ', '')

            
    #         try:
    #             # Assuming you have a connection pool or a single connection to your database
    #             conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
                
    #             # Insert data into the market_data table
    #             await conn.execute('''
    #                 INSERT INTO sec_filings(ticker, title, link, insertion_timestamp)
    #                 VALUES($1, $2, $3, NOW())
    #                 ''', ticker, title, link)

    #         except UniqueViolationError:
    #             print(f'Skipping')

    #     # if message.channel.id in RSS_FEEDS:
    #     #     print(message.content)

    #     if message.channel.id in CRYPTO:
    #         footer = message.embeds[0].footer.text
    #         footer = footer.split("|")

    #         ticker = footer[0].replace(' ','')
    #         side = footer[1].replace(' ','')
    #         dollar_cost = float(footer[2].replace(' ',''))
    #         timestamp = embeds[0].timestamp
    #         timestamp = str(timestamp.astimezone()).split('.')[0]
    #         # Convert string to datetime object
    #         timestamp_dt = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
    #         try:
    #             # Assuming you have a connection pool or a single connection to your database
    #             conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
                
    #             # Insert data into the market_data table
    #             await conn.execute('''
    #                 INSERT INTO crypto(ticker, side, dollar_cost, timestamp)
    #                 VALUES($1, $2, $3, $4)
    #                 ''', ticker, side, dollar_cost, timestamp_dt)
                
    #             # Close the connection if you're not using a connection pool
    #             await conn.close()
    #         except UniqueViolationError:
    #             print(f'Skipping')

    #     # if message.channel.id in RSI_FEEDS:
    #     #     footer = message.embeds[0].footer.text
    #     #     print(footer)

    #     if message.channel.id in MOMENTUM_SCALPS:
    #         footer = message.embeds[0].footer.text
    #         footer = footer.split("|")

    #         ticker = footer[0].replace(' ', '')
    #         timeframe = footer[1].replace(' ', '')
    #         move = footer[2].replace(' ', '')
    #         # Assuming you have a connection pool or a single connection to your database
    #         try:
    #             conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
                
    #             # Insert data into the market_data table
    #             await conn.execute('''
    #                 INSERT INTO momentum_scalps(ticker, timeframe, move, insertion_timestamp)
    #                 VALUES($1, $2, $3, NOW())
    #                 ''', ticker.replace(' ',''), timeframe.replace(' ', ''), move.replace(' ',''))
                
    #             # Close the connection if you're not using a connection pool
    #             await conn.close()
    #         except UniqueViolationError:
    #             pass

    #     if message.channel.id in opt_vol_ids:
    #         footer = message.embeds[0].footer.text
    #         timestamp = embeds[0].timestamp
    #         timestamp = str(timestamp.astimezone()).split('.')[0]
    #         timestamp_dt = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
    #         print(footer, timestamp_dt)

        


    #     if message.channel.id in SECTOR_DARKPOOLS:
    #         embeds = message.embeds
    #         title = embeds[0].title
    #         desc = embeds[0].description

    #         fields = embeds[0].fields

    #         names = [i.name for i in fields]
    #         values = [i.value for i in fields]
            
    

    #         ticker = title.split(': ')
    #         ticker = ticker[1]
    #         timestamp = now.strftime('%Y-%m-%d %H:%M:%S')
    #         size_price = values[1].split(' @ ')
    #         size = size_price[0].replace(',','')
    #         price = size_price[1].replace('$','')
    #         notional_value = values[2].split(' ')
    #         notional_value=notional_value[0].replace('$', '').replace(',','')
    #         channel = message.channel.name
    #         # Regex pattern to match emojis
    #         emoji_pattern = re.compile("["
    #                                 u"\U0001F600-\U0001F64F"  # emoticons
    #                                 u"\U0001F300-\U0001F5FF"  # symbols & pictographs
    #                                 u"\U0001F680-\U0001F6FF"  # transport & map symbols
    #                                 u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
    #                                 u"\U00002702-\U000027B0"
    #                                 u"\U000024C2-\U0001F251"
    #                                 "]+", flags=re.UNICODE)

    #         # Remove emojis from the string
    #         cleaned_string = emoji_pattern.sub(r'', channel)

    #         try:
    #             conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
                
    #             # Insert data into the market_data table
    #             await conn.execute('''
    #                 INSERT INTO dark_pools(sector, ticker, time, size, price, notional_value)
    #                 VALUES($1, $2, $3, $4, $5, $6)
    #                 ''', cleaned_string, ticker, timestamp, size, price, notional_value)
                
    #             # Close the connection if you're not using a connection pool
    #             await conn.close()
    #         except UniqueViolationError:
    #             pass

    #     if message.channel.id in STOCK_FEEDS:
    #         footer = embeds[0].footer.text
    #         footer = footer.split('|')
    #         ticker = footer[0]
    #         type = footer[1].replace('!','').replace(' ','').lower()
    #         price = footer[2]
    #         fifty_high = footer[3]
    #         fifty_low = footer[4]

    #         timestamp = embeds[0].timestamp
    #         timestamp = str(timestamp.astimezone()).split('.')[0]
    #         # Convert string to datetime object
    #         timestamp_dt = datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')

    #         print(type, price, fifty_high, fifty_low)

    #         if 'High' or 'Low' in title:
    #             try:
    #                 conn = await asyncpg.connect('postgresql://postgres:fud@localhost/polygon')
                    
    #                 # Insert data into the market_data table
    #                 await conn.execute('''
    #                     INSERT INTO new_prices(ticker, type, fifty_low, price, fifty_high, timestamp)
    #                     VALUES($1, $2, $3, $4, $5, $6)
    #                     ''', ticker, type, fifty_high, price, fifty_low, timestamp)
                    
    #                 # Close the connection if you're not using a connection pool
    #                 await conn.close()
    #             except UniqueViolationError:
    #                 pass


    #     if message.channel.id in OPTIONS_CONDITIONS:
    #         footer = embeds[0].footer.text
    #         print(footer)
    # await bot.process_commands(message)
# This dictionary will hold the conversation state for each user
conversations = {}
client = OpenAI(api_key=os.environ.get('YOUR_OPENAI_KEY'))
@bot.slash_command()
async def option_data(inter: disnake.ApplicationCommandInteraction):
    modal = OptionModal()
    await inter.response.send_modal(modal)


async def handle_interaction(interaction):
    # Call your async function here
    await server(interaction)
    # Edit the original message after handling the interaction
    await interaction.edit_original_message(content="New content here")
    

@bot.slash_command()
async def server(inter:disnake.AppCmdInter):
    """Gets server feeds"""
    await inter.response.defer()
    db_pool = await asyncpg.create_pool(database='polygon', user='postgres', password='fud')
    counter = 0
    while True:
        counter = counter + 1
        td9_query = "SELECT distinct ticker, timespan, insertion_timestamp FROM market_data order by insertion_timestamp DESC LIMIT 10;"  # Replace 
        sec_query = "SELECT distinct ticker, title, insertion_timestamp FROM sec_filings order by insertion_timestamp DESC limit 1;"
        momentum_query = "SELECT distinct ticker, timeframe, move, insertion_timestamp FROM momentum_scalps order by insertion_timestamp DESC limit 4;"
        reddit_query = "SELECT distinct subreddit, title, insertion_timestamp FROM reddit_posts order by insertion_timestamp DESC limit 4;"
        flow_query = f"SELECT ticker, strike, call_put, expiry, sentiment FROM ( SELECT ticker, strike, call_put, expiry, sentiment, timestamp, ROW_NUMBER() OVER (PARTITION BY ticker, strike, call_put, expiry, sentiment ORDER BY timestamp) as rn FROM flow ) subquery WHERE rn = 1 ORDER BY timestamp DESC limit 3;"
        rsi_query = f"SELECT ticker, timespan, status FROM ( SELECT ticker, timespan, status, timestamp, ROW_NUMBER() OVER (PARTITION BY ticker, timespan, status ORDER BY timestamp DESC) as rn FROM rsi_status ) subquery WHERE rn = 1 ORDER BY timestamp DESC LIMIT 10;"
        dark_pool_query = f"SELECT ticker, price, sector FROM ( SELECT ticker, price, sector, time, ROW_NUMBER() OVER (PARTITION BY ticker, price, sector ORDER BY time DESC) as rn FROM dark_pools ) subquery WHERE rn = 1 ORDER BY time DESC LIMIT 5;"
        new_prices_query = f"SELECT ticker, fifty_high, price, fifty_low FROM ( SELECT ticker, fifty_high, price, fifty_low, timestamp, ROW_NUMBER() OVER (PARTITION BY ticker ORDER BY timestamp DESC) as rn FROM new_prices ) subquery WHERE rn = 1 ORDER BY timestamp DESC LIMIT 5;"
        message_query = f"SELECT username, message FROM ( SELECT username, message, timestamp, ROW_NUMBER () OVER (PARTITION BY message ORDER BY timestamp DESC) as rn FROM messages ) subquery WHERE rn = 1 ORDER BY timestamp DESC LIMIT 5;"
        async with db_pool.acquire() as conn:
            td9_rows = await conn.fetch(td9_query)
            sec_rows = await conn.fetch(sec_query)
            momentum_rows = await conn.fetch(momentum_query)
            reddit_rows = await conn.fetch(reddit_query)
            flow_rows = await conn.fetch(flow_query)
            rsi_rows = await conn.fetch(rsi_query)
            dark_pool_rows = await conn.fetch(dark_pool_query)

            new_prices_rows = await conn.fetch(new_prices_query)

            messages_rows = await conn.fetch(message_query)

            messages_data = [[row[0], row[1][:25]] for row in messages_rows]

            td9_data = [[row[0], row[1], format_timestamp(row[2])] for row in td9_rows]
            sec_data = [[row[0], row[1][:15], format_timestamp(row[2])] for row in sec_rows]
            momentum_data = [[row[0], row[1], row[2], format_timestamp(row[3])] for row in momentum_rows]
            reddit_data = [[row[0][:25], row[1][:25], format_timestamp(row[2])] for row in reddit_rows]
            

            # Create tables using tabulate
            td9_table = tabulate(td9_data, headers=['Ticker', 'Date', 'Date'], tablefmt="fancy", showindex=False)
            sec_table = tabulate(sec_data, headers=['Ticker', 'Title', 'Date'], tablefmt="fancy", showindex=False)
            momentum_table = tabulate(momentum_data, headers=['Ticker', 'Frame', 'Move', 'Date'], tablefmt="fancy", showindex=False)
            reddit_table = tabulate(reddit_data, headers=['Channel', 'Title', 'Timestamp'], tablefmt="fancy", showindex=False)
            new_price_table = tabulate(new_prices_rows, headers='keys', tablefmt='fancy', showindex=False)

            messages_table = tabulate(messages_data, headers='keys', tablefmt='fancy', showindex=False)
            flow_table = tabulate(flow_rows, headers='keys', tablefmt='fancy', showindex=False)
            rsi_table = tabulate(rsi_rows, headers='keys', tablefmt='fancy', showindex=False)
            dark_pool_table = tabulate(dark_pool_rows, headers='keys', tablefmt='fancy', showindex=False)

            embed = disnake.Embed(title=f"GPT4-Turbo", description=f"```Viewing Live FUDSTOP Feeds```", color=disnake.Colour.dark_orange())
            embed.add_field(name=f"TD9s:", value=f"```py\n{td9_table}```", inline=False)
            
            embed.add_field(name=f"Scalps:", value=f"```py\n{momentum_table}```", inline=False)
            embed.add_field(name=f"Opening Flow:", value=f"```py\n{flow_table}```", inline=False)
            embed.add_field(name=f"RSI:", value=f"```py\n{rsi_table}```", inline=False)
            embed.add_field(name=f"Dark Pools:", value=f"```py\n{dark_pool_table}```", inline=False)
            embed.add_field(name=f"New Prices:", value=f"```py\n{new_price_table}```", inline=False)
            embed.add_field(name=f"SEC Filings:", value=f"```py\n{sec_table}```", inline=False)
            embed.add_field(name=f'Reddit:', value=f"```py\n{reddit_table}```", inline=False)
            embed.add_field(name=f"Messages:", value=f"```py\n{messages_table}```", inline=False)
            await inter.edit_original_message(embed=embed)
            button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, emoji="<a:_:1104142591110418585>")
            button.callback = handle_interaction
            view=disnake.ui.View()
            if counter == 400:
                view.add_item(button)
                await inter.edit_original_message(view=view)
                break

# @bot.slash_command()
# async def options_database(inter:disnake.ApplicationCommandInteraction):
#     """Use a Modal to query the database for options."""
#     modal = SQLQueryModal()
#     await inter.response.send_modal(modal)


@bot.slash_command()
async def screener(inter: disnake.AppCmdInter,
                   ask_gte: str = None, ask_lte: str = None,
                   bid_gte: str = None, bid_lte: str = None,
                   changeratio_gte: str = None, changeratio_lte: str = None,
                   close_gte: str = None, close_lte: str = None,
                   delta_gte: str = None, delta_lte: str = None,
                   direction: str = None,  # This might need to be handled differently since it's a list
                   expiredate_gte: str = None, expiredate_lte: str = None,
                   gamma_gte: str = None, gamma_lte: str = None,
                   implvol_gte: str = None, implvol_lte: str = None,
                   openinterest_gte: str = None, openinterest_lte: str = None,
                   theta_gte: str = None, theta_lte: str = None,
                   volume_gte: str = None, volume_lte: str = None):
    await inter.response.defer()

    # Creating an instance of WebulScreener
    screener = WebulScreener()

    # Example usage of the query method with parameters from the command
    query_result = screener.query(
        ask_gte=ask_gte, ask_lte=ask_lte,
        bid_gte=bid_gte, bid_lte=bid_lte,
        changeRatio_gte=changeratio_gte, changeRatio_lte=changeratio_lte,
        close_gte=close_gte, close_lte=close_lte,
        delta_gte=delta_gte, delta_lte=delta_lte,
        direction=[direction] if direction else None,  # Handling direction as a list
        expireDate_gte=expiredate_gte, expireDate_lte=expiredate_lte,
        gamma_gte=gamma_gte, gamma_lte=gamma_lte,
        implVol_gte=implvol_gte, implVol_lte=implvol_lte,
        openInterest_gte=openinterest_gte, openInterest_lte=openinterest_lte,
        theta_gte=theta_gte, theta_lte=theta_lte,
        volume_gte=volume_gte, volume_lte=volume_lte
    )
    query_result_df = pd.DataFrame(query_result)
    query_result_df = query_result_df.drop(columns=['id'])
    chunks = [query_result_df[i:i + 3860] for i in range(0, len(query_result_df), 3860)]
    
    embeds =[]
    for chunk in chunks:
    
        embed = disnake.Embed(title=f"Screener Results:", description=f"```py\n{chunk}```")
        # Here, handle the query_result as needed, e.g., sending a message
        view = disnake.ui.View()
        ids = query_result.get('id')
        symbols = query_result.get('symbol')
        strikes = query_result.get('strike')
        expiry = query_result.get('expiry')
        call_put = query_result.get('call_put')
        embed.set_footer(text=f'Implemented by FUDSTOP')
        embeds.append(embed)
    view.add_item(ScreenerSelect(query_result))
    await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(ScreenerSelect(query_result)))
        


@bot.command()
async def gpt4(ctx):
    # Start a new conversation with the user
    conversations[ctx.author.id] = []
    
    # Send an initial message to the user
    await ctx.send("> # Slave Bot\n> Online. \n\n> Your work is my...work.. Type to chat... or.. type stop to quit.")

    while True:
        # Wait for a message from the same user
        message = await bot.wait_for(
            "message",
            check=lambda m: m.author == ctx.author and m.channel == ctx.channel
        )

        # Check if the user wants to stop the conversation
        if message.content.lower() == "stop":
            await ctx.send("Goodbye! If you need help again, just call me.")
            del conversations[ctx.author.id]  # Clean up the conversation
            break

        # Append the user's message to the conversation
        conversations[ctx.author.id].append({"role": "user", "content": message.content + "YOU ARE ONLY TO REPLY IN CODE. CODE ONLY. NO MARKDOWN. ONLY CODE!"})

        # Send the conversation to OpenAI
        response = client.chat.completions.create(
            model="gpt-4-1106-preview",
            messages=conversations[ctx.author.id],
            temperature=0.33,
            max_tokens=4096        )

        # Get the content from the OpenAI response
        ai_response = response.choices[0].message.content

        # Append the AI's response to the conversation
        conversations[ctx.author.id].append({"role": "assistant", "content": ai_response})
        embeds = []
        chunks = [ai_response[i:i + 3860] for i in range(0, len(ai_response), 3860)]
        for chunk in chunks:
            embed = disnake.Embed(title=f"GPT4-Turbo", description=f"{chunk}", color=disnake.Colour.dark_orange())
            embed.add_field(name=f"Your prompt:", value=f"> **{message.content[:400]}**", inline=False)
            embeds.append(embed)

        await ctx.send(embed=embeds[0], view=AlertMenus(embeds))



# @bot.command()
# async def cfr(ctx, *, query: str):
#     results = fetch_results(query)
    
#     async def create_table(conn):
#         await conn.execute('''
#             CREATE TABLE IF NOT EXISTS cfr_data (
#                 starts_on TEXT,
#                 ends_on TEXT,
#                 type TEXT,
#                 hierarchy_title TEXT,
#                 hierarchy_subtitle TEXT,
#                 hierarchy_chapter TEXT,
#                 hierarchy_subchapter TEXT,
#                 hierarchy_part TEXT,
#                 hierarchy_subpart TEXT,
#                 hierarchy_subject_group TEXT,
#                 hierarchy_section TEXT,
#                 hierarchy_appendix TEXT,
#                 hierarchy_headings_title TEXT,
#                 hierarchy_headings_subtitle TEXT,
#                 hierarchy_headings_chapter TEXT,
#                 hierarchy_headings_subchapter TEXT,
#                 hierarchy_headings_part TEXT,
#                 hierarchy_headings_subpart TEXT,
#                 hierarchy_headings_subject_group TEXT,
#                 hierarchy_headings_section TEXT,
#                 hierarchy_headings_appendix TEXT,
#                 headings_title TEXT,
#                 headings_subtitle TEXT,
#                 headings_chapter TEXT,
#                 headings_subchapter TEXT,
#                 headings_part TEXT,
#                 headings_subpart TEXT,
#                 headings_subject_group TEXT,
#                 headings_section TEXT,
#                 headings_appendix TEXT,
#                 full_text_excerpt TEXT,
#                 score REAL,
#                 structure_index INT,
#                 reserved BOOLEAN,
#                 removed BOOLEAN,
#                 change_types_effective_cross_reference TEXT,
#                 change_types_cross_reference TEXT,
#                 change_types_effective TEXT,
#                 change_types_initial TEXT
#             );
#         ''')

#     async def insert_data(conn, results):
#         insert_query = '''
#             INSERT INTO cfr_data (
#                 starts_on, ends_on, type, hierarchy_title, hierarchy_subtitle,
#                 hierarchy_chapter, hierarchy_subchapter, hierarchy_part,
#                 hierarchy_subpart, hierarchy_subject_group, hierarchy_section,
#                 hierarchy_appendix, hierarchy_headings_title,
#                 hierarchy_headings_subtitle, hierarchy_headings_chapter,
#                 hierarchy_headings_subchapter, hierarchy_headings_part,
#                 hierarchy_headings_subpart, hierarchy_headings_subject_group,
#                 hierarchy_headings_section
#             ) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $30, $31, $32, $33, $34, $35, $36, $37, $38);
#         '''
#         await conn.executemany(insert_query, results)
    
#     if 'results' in results:

#         df = pd.DataFrame(results['results'])
#         csv_buffer = io.StringIO()
#         df.to_csv(csv_buffer, index=False)
#         csv_buffer.seek(0)
#         await ctx.send(file=disnake.File(fp=csv_buffer, filename='cfr_results.csv'))
    

from apis.oic.oic_sdk import OICSDK
oic_sdk = OICSDK()

@bot.command()
async def monitor(ctx, ticker):
    """Monitors an option in great detail"""
    try:
        df = await oic_sdk.options_monitor(ticker=ticker)

        df.as_dataframe.to_csv('data/oic/options_monitor.csv', index=False)
        
        await ctx.send(file=disnake.File('data/oic/options_monitor.csv'))
    except CommandInvokeError:
        await bot.restar


@bot.command()
async def active(ctx):
    """Returns the most active options from the OIC"""

  


    df = oic_sdk.most_active_options()

    df.to_csv('data/oic/most_active_options.csv', index=False)


    await ctx.send(file=disnake.File('data/oic/most_active_options.csv'))


    


        # # Validate tickers
        # valid_tickers = [ticker for ticker in tickers if ticker in self.valid_tickers]
        # invalid_tickers = list(set(tickers) - set(valid_tickers))

        # if not valid_tickers:
        #     # No valid tickers were entered
        #     await interaction.response.send_message(
        #         "None of the entered tickers were recognized. Please try again with valid ticker symbols.",
        #         ephemeral=True
        #     )
        #     return

        # # If there are invalid tickers, inform the user but proceed with valid ones
        # if invalid_tickers:
        #     await interaction.followup.send(
        #         f"The following tickers were not recognized and will be ignored: {', '.join(invalid_tickers)}",
        #         ephemeral=True
        #     )

        # # Proceed with scanning the valid tickers
        # results = await scan_bars(valid_tickers, timeframe)
        # await interaction.followup.send(results)

from _markets.stock_market import StockMarketLive




# stock_market = StockMarketLive()
# @bot.slash_command()
# async def stream(inter: disnake.AppCmdInter, ticker:str):
#     """Stream live trades for a ticker."""
#     await inter.response.defer()
    
#     await stock_market.connect()
#     counter = 0
#     while True:
#         counter = counter + 1
#         data = await stock_market.fetch_latest_trade(ticker)
#         if data:
#             # Format the message with the trade data
#             message = f"> # Latest trade for {ticker} | Price: ${data['price']} | Size: {data['size']} | Time: {data['timestamp']}"
#         else:
#             # No trade data found
#             message = f"> No recent trades found for {ticker}."
#         await inter.edit_original_message(f"> # {message}")

#         if counter == 250:
#             await inter.send(f'> # Stream ended.')
#             break
    
@bot.command()
async def cal(ctx):
    # Set up headless browser options for Selenium
    chrome_options = Options()
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--disable-gpu")
    chrome_options.add_argument("--window-size=1920x1080")

    # Initialize the Chrome driver
    driver = webdriver.Chrome(options=chrome_options)

    try:
        # Get the webpage snapshot
        driver.get("https://www.newyorkfed.org/research/calendars/nationalecon_cal")

        # Get today's date in the format that matches the calendar on the webpage
        today = datetime.now().strftime("%d")  # Format to match the date format on the calendar, e.g. "13" for 13th
        if today.startswith('0'):
            # Remove leading zero for single digit dates if necessary
            today = today[1:]

        # Wait for the calendar element that matches today's date to be present in the DOM
        wait = WebDriverWait(driver, 10)
        # Replace 'dateElementLocator' with the actual locator that matches the date on the calendar
        date_element_locator = f"//someElement[contains(text(), '{today}')]"
        date_element = wait.until(EC.presence_of_element_located((By.XPATH, date_element_locator)))

        # Scroll the date element into view
        driver.execute_script("arguments[0].scrollIntoView({block: 'center'});", date_element)

        # Optionally, wait for any animations or dynamic content to settle
        driver.implicitly_wait(2)

        # Take a screenshot of the visible part of the page, hopefully centered on today's date
        screenshot = driver.get_screenshot_as_png()

        # Convert the screenshot to a Discord file-like object
        screenshot_file = io.BytesIO(screenshot)
        screenshot_file.seek(0)
        screenshot_file.name = 'calendar.png'

        # Create a disnake File object
        discord_file = disnake.File(screenshot_file, filename='calendar.png')

        # Create embed to attach the image
        embed = disnake.Embed(title="Economic Calendar", color=0x1a1a1a)
        embed.set_image(url="attachment://calendar.png")

        # Send the embed with the screenshot in the current channel
        await ctx.send(embed=embed, file=discord_file)

    finally:
        # Make sure to quit the driver to free up resources
        driver.quit()

timeframe_choices = [
    disnake.OptionChoice(name="1 Minute", value="m1"),
    disnake.OptionChoice(name="5 Minutes", value="m5"),
    disnake.OptionChoice(name="10 Minutes", value="m10"),
    disnake.OptionChoice(name="15 Minutes", value="m15"),
    disnake.OptionChoice(name="30 Minutes", value="m30"),
    disnake.OptionChoice(name="1 Hour", value="m60"),
    disnake.OptionChoice(name="2 Hours", value="m120"),
    disnake.OptionChoice(name="4 Hours", value="m240"),
    disnake.OptionChoice(name="Daily", value="d1"),
    disnake.OptionChoice(name="Weekly", value="w")
]





async def fetch_options():
    conn = await asyncpg.connect(user=os.environ.get('DB_USER'), password=os.environ.get('DB_PASSWORD'), database=os.environ.get('DB_NAME'), host=os.environ.get('DB_HOST'))
    rows = await conn.fetch("""WITH LatestPrices AS (
    SELECT 
        a.ticker, 
        a.strike, 
        a.expiry, 
        a.call_put, 
        a.theta,
        b.low AS current_price,
        a.bid, 
        a.ask,
        ROW_NUMBER() OVER (PARTITION BY a.ticker, a.strike, a.expiry, a.call_put ORDER BY b.timestamp DESC) AS rn
    FROM 
        options_data a
    INNER JOIN 
        option_aggs b ON a.ticker = b.ticker
                      AND a.strike = b.strike
                      AND a.expiry = b.expiry
                      AND a.call_put = b.call_put
    WHERE 
        a.bid >= 0.14 AND a.ask <= 1.00
        AND a.theta >= -0.02
),
AllTimeLows AS (
    SELECT 
        a.ticker, 
        a.strike, 
        a.expiry, 
        a.call_put, 
        MIN(b.low) AS all_time_low
    FROM 
        options_data a
    INNER JOIN 
        option_aggs b ON a.ticker = b.ticker
                      AND a.strike = b.strike
                      AND a.expiry = b.expiry
                      AND a.call_put = b.call_put
    GROUP BY 
        a.ticker, a.strike, a.expiry, a.call_put
)
SELECT 
    l.ticker, 
    l.strike, 
    l.expiry, 
    l.call_put, 
    l.current_price
FROM 
    LatestPrices l
INNER JOIN 
    AllTimeLows a ON l.ticker = a.ticker
                 AND l.strike = a.strike
                 AND l.expiry = a.expiry
                 AND l.call_put = a.call_put
                 AND l.current_price = a.all_time_low
WHERE 
    l.rn = 1;""")
    await conn.close()
    return rows

@commands.is_owner()
@bot.command()
async def clear(ctx: commands.Context, limit: int = 10):
    await ctx.channel.purge(limit=limit)

pytrends = TrendReq(hl='en-US', tz=360)
@bot.command()
async def trends(ctx):


    
    trending_searches_df = pytrends.trending_searches(pn='united_states')  # you can change the region
    embed = disnake.Embed(title=f"Current Trends on Google:", description=f"```py\n{trending_searches_df}```", color=disnake.Colour.dark_green())
    await ctx.send(embed=embed)





@bot.command()
async def iv(ctx, ticker, strike, call_put, expiry):
    """Gets IV for a ticker"""
    
 
    await opts.connect()
    async for results in opts.fetch_iter(query=f"""select option_symbol from options_data where ticker = '{ticker}' and strike = {strike} and expiry = '{expiry}' and call_put = '{call_put}';"""):
        view = disnake.ui.View()
        option_symbol = results[0]
        data = []
        class Select(disnake.ui.Select):
            def __init__(self):
                self.option_symbol = option_symbol
                super().__init__( 
                    placeholder='Select -->',
                    min_values=1,
                    max_values=1,
                    options= [disnake.SelectOption(label=f'{option_symbol}')]
                )

            async def callback(self, inter:disnake.AppCmdInter):
                while True:
                    await inter.response.defer()

                    data = await opts.get_universal_snapshot(ticker=option_symbol)
                    df = pd.DataFrame(data)
                    # Selecting the 'iv' column
                    print(df.columns)
                    iv_column = df['IV']

                    await inter.edit_original_message(iv_column)
        view.add_item(Select())
        await ctx.send(view=view)

# @bot.command()
# async def puts(ctx, ticker: str):
#     """Gets at the money puts for a ticker"""
#     data = sdk.atm_puts(ticker)
#     data = pd.DataFrame(data)
#     filename = f'data/yf_atm_puts.csv'
#     data.to_csv(filename)
#     await ctx.send(file=disnake.File(filename))


@bot.command()
async def related(ctx, keyword):
    related_queries = pytrends.related_queries()
    # Define keywords
    keywords = [f"{keyword}"]

    # Build payload
    pytrends.build_payload(keywords, timeframe='today 12-m')

    # Fetch related queries
    related_queries = pytrends.related_queries()

    
    embed = disnake.Embed(title=f"Current Trends on Google:", description=f"```py\n{related_queries}```", color=disnake.Colour.dark_green())
    await ctx.send(embed=embed)






# Command to print webhook URLs for a specific channel
@bot.command(name='webhooks')
async def fetch_webhooks(ctx):
    # Ensure the author has the necessary permissions
    if ctx.author.guild_permissions.manage_webhooks:
        webhooks = await ctx.channel.webhooks()
        if webhooks:
            response = "\n".join([webhook.url for webhook in webhooks])
            await ctx.send(f"Webhook URLs for {ctx.channel.name}:\n{response}")
        else:
            await ctx.send("No webhooks found in this channel.")
    else:
        await ctx.send("You do not have permission to manage webhooks.")


@bot.slash_command()
async def study_greeks(inter:disnake.AppCmdInter, ticker:str):
    """Study greeks and their effects on price action."""

    await inter.response.defer()
    while True:
        price = yf.fast_info('AAPL')

        current_price = price[price[0] == 'lastPrice'][1].values[0]

        moenyness = await opts.get_option_strikes_by_moneyness(ticker=ticker, current_price=current_price, expiration_date='2023-12-01')

        embed = disnake.Embed(title=f"Greek Study", description=f"Studying Greeks for {ticker}")

        for term, categories in moenyness.items():
            embed.add_field(name=f"Term: {term}", value="-" * 10, inline=False)  # Separator for terms
            for moneyness_category, options in categories.items():
                # Assume each 'option' object has attributes like 'delta', 'gamma', etc.
                for option, category in options:
                    delta = getattr(option, 'delta', 'N/A')
                    gamma = getattr(option, 'gamma', 'N/A')
                    vega = getattr(option, 'vega', 'N/A')
                    theta = getattr(option, 'theta', 'N/A')
                    ask = getattr(option, 'ask', 'N/A')
                    bid = getattr(option, 'ask', 'N/A')
                    strike = getattr(option, 'strike', 'N/A')
                    expiry = getattr(option, 'expiry', 'N/A') 

                    option_info = f"> Delta: {round(float(delta),4)}\n> Gamma: {round(float(gamma),4)}\n> Vega: {round(float(vega),4)}\n> Theta: {round(float(theta),2)}\n> Bid: **${bid}**\n> Ask: **${ask}**"
                    embed.add_field(name=f"${ticker} | {strike} | {expiry}", value=option_info, inline=True)

        await inter.edit_original_message(embed=embed)


@bot.command()
async def pricecheck(ctx, ticker:str):


    result = await opts.get_theoretical_price(ticker)
    
    df = pd.DataFrame(result)
    df = df.sort_values('theoretical_price', ascending=False)
    df.to_csv('theo_check.csv')

    await ctx.send(file=disnake.File('theo_check.csv'))

opts = PolygonOptions()

async def find_extreme_tickers(pool):
    # SQL query to find tickers that are overbought or oversold on both day and week timespans
    query_sql = """
    SELECT day_rsi.ticker, day_rsi.status
    FROM rsi as day_rsi
    JOIN rsi as week_rsi ON day_rsi.ticker = week_rsi.ticker
    WHERE day_rsi.timespan = 'day' 
    AND week_rsi.timespan = 'week'
    AND day_rsi.status IN ('overbought', 'oversold')
    AND week_rsi.status IN ('overbought', 'oversold')
    AND day_rsi.status = week_rsi.status;
    """

        # Execute the query using the provided connection pool
    async with pool.acquire() as conn:
        records = await conn.fetch(query_sql)
        return [(record['ticker'], record['status']) for record in records]

async def find_plays():
    db_config = {
        'user': 'postgres',
        'password': 'fud',
        'database': 'opts',
        'host': '127.0.0.1',
        'port': 5432
    }

    async with asyncpg.create_pool(**db_config) as pool:
        extreme_tickers_with_status = await find_extreme_tickers(pool)

        # To separate the tickers and statuses, you can use list comprehension
        extreme_tickers = [ticker for ticker, status in extreme_tickers_with_status]
        statuses = [status for ticker, status in extreme_tickers_with_status]
        all_options_df_calls =[]
        all_options_df_puts = []
        for ticker, status in extreme_tickers_with_status:
            if status == 'overbought':
                print(f"Ticker {ticker} is overbought.")
                all_options = await opts.get_option_chain_all(ticker, expiration_date_gte='2024-03-01', expiration_date_lite='2024-06-30', contract_type='put')
                
                for i in range(len(all_options.theta)):  # Assuming all lists are of the same length
                    theta_value = all_options.theta[i]
                    volume = all_options.volume[i]
                    open_interest = all_options.open_interest[i]
                    ask = all_options.ask[i]
                    bid = all_options.bid[i]

                    # Conditions
                    theta_condition = theta_value is not None and theta_value >= -0.03
                    volume_condition = volume is not None and open_interest is not None and volume > open_interest
                    price_condition = ask is not None and bid is not None and 0.25 <= bid <= 1.75 and 0.25 <= ask <= 1.75

                    if theta_condition and volume_condition and price_condition:
                        df = pd.DataFrame([all_options.ticker, all_options.underlying_ticker, all_options.strike, all_options.contract_type, all_options.expiry])
                        all_options_df_puts.append(df)  #

            if status == 'oversold':
                print(f"Ticker {ticker} is oversold.")
                all_options = await opts.get_option_chain_all(ticker, expiration_date_gte='2024-03-01', expiration_date_lite='2024-11-30', contract_type='call')
                
                for i in range(len(all_options.theta)):  # Assuming all lists are of the same length
                    theta_value = all_options.theta[i]
                    volume = all_options.volume[i]
                    open_interest = all_options.open_interest[i]
                    ask = all_options.ask[i]
                    bid = all_options.bid[i]

                    # Conditions
                    theta_condition = theta_value is not None and theta_value >= -0.03
                    volume_condition = volume is not None and open_interest is not None and volume > open_interest
                    price_condition = ask is not None and bid is not None and 0.25 <= bid <= 1.75 and 0.25 <= ask <= 1.75

                    if theta_condition and volume_condition and price_condition:
                        # Assuming all_options.df is a DataFrame containing the current option data
                        df = pd.DataFrame([all_options.ticker, all_options.strike, all_options.contract_type, all_options.expiry])
                        all_options_df_calls.append(df)  #
        # Concatenate all the dataframes
        final_df_calls = pd.concat(all_options_df_calls, ignore_index=True)
        final_df_puts = pd.concat(all_options_df_puts, ignore_index=True)
        print(final_df_calls, final_df_puts)
        return final_df_calls, final_df_puts, extreme_tickers, statuses


class CallResults(disnake.ui.Select):
    def __init__(self, options: List[disnake.SelectOption], all_options:pd.DataFrame=None):
        self.all_options=all_options
        self.options = options
        super().__init__( 
            custom_id='callresults',
            min_values=1,
            max_values=1,
            placeholder='Results -->',
            options=options
        )


    async def callback(self, inter:disnake.AppCmdInter):

        if self.values[0] == self.values[0]:

            data = self.all_options.head(10)


            await inter.response.edit_message(data)

    
@bot.slash_command()
async def play(inter:disnake.AppCmdInter, type:str=commands.Param(choices=['calls','puts'])):
    await inter.response.defer()
    await inter.edit_original_message(f'Finding plays for {type}..')


    calls, puts, extreme_tickers, statuses = await find_plays()


    view = disnake.ui.View()
    if type == 'calls':
        options = []
        
        for ticker, status in zip(extreme_tickers, statuses):
            if status == 'oversold':
                
                options.append(disnake.SelectOption(label=ticker, description=status))
        await inter.edit_original_message(file=disnake.File('calls.csv'), view=view.add_item(CallResults(options,calls)))
    elif type == 'puts':
        options = []
        
        for ticker, status in zip(extreme_tickers, statuses):
            if status == 'overbought':
                options.append(disnake.SelectOption(label=ticker, description=status))
        await inter.edit_original_message(file=disnake.File('puts.csv'), view=view.add_item(CallResults(options,puts)))

    
# Specify the directory where your cogs are located
cogs_dir = "C:/users/chuck/fudstop/cogs"

print("Path being used:", cogs_dir)

# Get a list of all cog files in the specified directory
cog_files = [filename for filename in os.listdir(cogs_dir) if filename.endswith(".py")]

# Load each cog extension
for cog_file in cog_files:
    # Construct the full module name for the cog
    cog_name = f"cogs.{cog_file[:-3]}"  # Remove '.py' extension and prepend 'cogs.'

    try:
        # Load the cog extension
        bot.load_extension(cog_name)
        print(f"Loaded cog: {cog_name}")
    except Exception as e:
        print(f"Failed to load cog: {cog_name}\nError: {str(e)}")
bot.run(os.environ.get('RESEARCH_BOT'))


import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)
from dotenv import load_dotenv
load_dotenv()
from typing import Optional
import os
import sys
import pandas as pd
from pathlib import Path
import requests
from asyncio import Lock
from datetime import datetime, timedelta
from ..webull.webull_trading import WebullTrading
import numpy as np
from urllib.parse import unquote
lock = Lock()
trading = WebullTrading()
# Function to map Pandas dtypes to PostgreSQL types
def dtype_to_postgres(dtype):
    if pd.api.types.is_integer_dtype(dtype):
        return 'INTEGER'
    elif pd.api.types.is_float_dtype(dtype):
        return 'REAL'
    elif pd.api.types.is_bool_dtype(dtype):
        return 'BOOLEAN'
    elif pd.api.types.is_datetime64_any_dtype(dtype):
        return 'TIMESTAMP'
    elif pd.api.types.is_string_dtype(dtype):
        return 'TEXT'
    else:
        return 'TEXT'  # Default type
class PolygonSync:
    def __init__(self):
        self.connection_string = os.environ.get('POLYGON_STRING')

        self.database = os.environ.get('DB_NAME')

        self.api_key = os.environ.get('YOUR_POLYGON_KEY')
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.one_year_from_now = (datetime.now() + timedelta(days=365)).strftime('%Y-%m-%d')
        self.one_year_ago = (datetime.now() - timedelta(days=365)).strftime('%Y-%m-%d')



    def get_polygon_logo(self, symbol: str) -> Optional[str]:
            """
            Fetches the URL of the logo for the given stock symbol from Polygon.io.

            Args:
                symbol: A string representing the stock symbol to fetch the logo for.

            Returns:
                A string representing the URL of the logo for the given stock symbol, or None if no logo is found.

            Usage:
                To fetch the URL of the logo for a given stock symbol, you can call:
                ```
                symbol = "AAPL"
                logo_url = await sdk.get_polygon_logo(symbol)
                if logo_url is not None:
                    print(f"Logo URL: {logo_url}")
                else:
                    print(f"No logo found for symbol {symbol}")
                ```
            """
            url = f"https://api.polygon.io/v3/reference/tickers/{symbol}?apiKey={self.api_key}"
            data = requests.get(url).json()
                    
            if 'results' not in data:
                # No results found
                return None
            
            results = data['results']
            branding = results.get('branding')

            if branding and 'icon_url' in branding:
                encoded_url = branding['icon_url']
                decoded_url = unquote(encoded_url)
                url_with_api_key = f"{decoded_url}?apiKey={self.api_key}"
                return url_with_api_key


from fudstop.apis.webull.webull_screener import WebulScreener


screener = WebulScreener()



import asyncio
from fudstop.apis.webull.webull_options import WebullOptionsData
import requests

async def main():
    r = requests.get("https://quotes-gw.webullfintech.com/api/quote/option/quotes/queryBatch?derivativeIds=1040230807,1040231298,1040230811,1040231397,1040231421,1040230589,1040231323,1040230617,1040230680,1040230731,1040230704,1040230491,1040229914,1040230151,1040230172,1040231333,1040230662", headers=screener.headers).json()
   
    
    data = WebullOptionsData(r)


    print(data.askVolume)
    


asyncio.run(main())





import yfinance as yf
from _markets.list_sets.ticker_lists import all_tickers
from datetime import datetime
import pandas as pd
from ..helpers import format_large_numbers_in_dataframe
import numpy as np
# Set options to display all rows and columns
pd.set_option('display.max_rows', None)  # None means show all rows


class yfSDK:
    def __init__(self):
        self.tickers = all_tickers
  
    def batch_insert_dataframe(self, df, table_name, unique_columns, batch_size=250):
        with self.lock:
            if not self.table_exists(table_name):
                self.create_table(df, table_name, unique_columns)

            df = df.copy()
            df.dropna(inplace=True)
            df['insertion_timestamp'] = [datetime.now() for _ in range(len(df))]

            records = df.to_records(index=False)
            data = list(records)

            with self.pool.acquire() as connection:
                column_types = connection.fetch(
                    f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'"
                )
                type_mapping = {col: next((item['data_type'] for item in column_types if item['column_name'] == col), None) for col in df.columns}

                with connection.transaction():
                    insert_query = f"""
                    INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                    VALUES ({', '.join('%s' for _ in df.columns)})
                    ON CONFLICT ({unique_columns})
                    DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                    """

                    batch_data = []
                    for record in data:
                        new_record = []
                        for col, val in zip(df.columns, record):
                            pg_type = type_mapping[col]

                            if val is None:
                                new_record.append(None)
                            elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))
                            elif isinstance(val, datetime):
                                new_record.append(pd.Timestamp(val).to_pydatetime())
                            elif pg_type in ['double precision', 'real'] and not isinstance(val, str):
                                new_record.append(float(val))
                            elif isinstance(val, np.int64):
                                new_record.append(int(val))
                            elif pg_type == 'integer' and not isinstance(val, int):
                                new_record.append(int(val))
                            else:
                                new_record.append(val)

                        batch_data.append(new_record)

                        if len(batch_data) == batch_size:
                            try:
                                connection.executemany(insert_query, batch_data)
                                batch_data.clear()
                            except Exception as e:
                                print(f"An error occurred while inserting the record: {e}")
                                connection.execute('ROLLBACK')
                                raise

    def save_to_history(self, df, main_table_name, history_table_name):
        # Assume the DataFrame `df` contains the records to be archived
        if not self.table_exists(history_table_name):
            self.create_table(df, history_table_name, None)

        df['archived_at'] = datetime.now()  # Add an 'archived_at' timestamp
        self.batch_insert_dataframe(df, history_table_name, None)

    def table_exists(self, table_name):
        query = f"SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = '{table_name}');"

        with self.pool.acquire() as conn:
            with conn.transaction():
                exists = conn.fetchval(query)
        return exists
    def balance_sheet(self, ticker:str, frequency:str='quarterly', pretty:bool=None, as_dict:bool=False):
        """
        Gets balance sheet information for a ticker.

        Arguments:

        >>> Frequency: The frequency. quarterly / annual (default quarterly)

        >>> As Dict: bool - return as a dictionary (optional - default FALSE)

        >>> Pretty: (optional - pretty prent)

        """

        

        data = yf.Ticker(ticker)
        if pretty == None:
            pretty = False

        balance_sheet = data.get_balance_sheet(freq=frequency,pretty=pretty, as_dict=as_dict)
        formatted_data = format_large_numbers_in_dataframe(balance_sheet)
        return formatted_data 

    

    def get_cash_flow(self, ticker:str, frequency:str='quarterly', pretty:bool=False, as_dict:bool=False):
        """
        Gets cash flow information for a ticker.

        Arguments:

        >>> Frequency: The frequency. quarterly / annual (default quarterly)

        >>> As Dict: bool - return as a dictionary (optional - default FALSE)

        >>> Pretty: (optional - pretty prent)
        """
        data = yf.Ticker(ticker).get_cash_flow(freq=frequency,pretty=pretty, as_dict=as_dict)


        formatted_data = format_large_numbers_in_dataframe(data)
        return formatted_data





    def get_all_candles(self, tickers:str):
        """
        Gets OHLC, adj.Close and Volume data for ALL DATES

        Arguments:


        >>> Tickers: a list of comma separated tickers. (default ALL TICKERS)

        >>> Period: the period to gather data for. OPTIONS = 
        
            >>> 1m,2m,5m,15m,30m,60m,90m,1h,1d,5d,1wk,1mo,3mo 
            
            
            Intraday data cannot extend last 60 days
        
        """


        try:
            chart_data = yf.download(tickers)
        except Exception as e:
            print(f'Error processing data. - {e}')



            chart_data = pd.DataFrame(chart_data).reset_index(None)


            return chart_data
        
    

    def dividends(self, ticker:str):
        """
        Returns historic dividends for a ticker
        
        """
        try:
            data = yf.Ticker(ticker).get_dividends()

            return data
        except Exception as e:
            return(f"No dividends found for {ticker}. {e}")
        

    def fast_info(self, ticker:str):
        """
        Arguments:

        >>> Limit: the number of results to return (optional - default 15)
        """
        data = yf.Ticker(ticker=ticker).get_fast_info().items()


        


        df = pd.DataFrame(data)
        df.reset_index(drop=True, inplace=True)
        formatted_data = format_large_numbers_in_dataframe(df)
        return formatted_data 



    def financials(self, ticker:str, frequency:str='quarterly', as_dict:bool=False, pretty:bool=False):

        """
        Gets all financials for a ticker.


        Arguments:

        >>> Frequency: The frequency. quarterly / annual (default quarterly)

        >>> As Dict: bool - return as a dictionary (optional - default FALSE)

        >>> Pretty: (optional - pretty prent)
        """
        data = yf.Ticker(ticker=ticker).get_financials(freq=frequency,as_dict=as_dict, pretty=pretty)

        formatted_data = format_large_numbers_in_dataframe(data)
        return formatted_data 
    

    def income_statement(self, ticker:str, frequency:str='quarterly', as_dict:bool=False, pretty:bool=False):
        """
        Gets the income statement for a ticker.

        Arguments:

        >>> Frequency: The frequency. quarterly / annual (default quarterly)

        >>> As Dict: bool - return as a dictionary (optional - default FALSE)

        >>> Pretty: (optional - pretty prent)

        """
        data = yf.Ticker(ticker=ticker).get_income_stmt(freq=frequency,as_dict=as_dict,pretty=pretty)


        formatted_data = format_large_numbers_in_dataframe(data)
        return formatted_data     
    

    def get_info(self, ticker:str):
        """
        Returns a large dictionary of information for a ticker.

        Arguments:

        None
        
        """
        data  = yf.Ticker(ticker).get_info().items()

        df = pd.DataFrame(data)
        formatted_data = format_large_numbers_in_dataframe(df)
        return formatted_data     

    


    def institutional_holdings(self, ticker:str):
        """
        Gets institutional holdoings.

        Arguments:


        
        """

        data =yf.Ticker(ticker).get_institutional_holders()

 

        formatted_data = format_large_numbers_in_dataframe(data)

        # Convert the '% Out' column to float (if it's not already)
        formatted_data['% Out'] = formatted_data['% Out'].astype(float)

        # Round the '% Out' column to 3 decimal places
        formatted_data['% Out'] = formatted_data['% Out'].round(3)
        formatted_data.set_index('Date Reported', inplace=True)

        return formatted_data


    def mutual_fund_holders(self, ticker:str):
        """
        Gets mutual fund holders


        Arguments:

        >>> 
        
        """

        data = yf.Ticker(ticker=ticker).get_mutualfund_holders()

        formatted_data = format_large_numbers_in_dataframe(data)
        # Convert the '% Out' column to float (if it's not already)
        formatted_data['% Out'] = formatted_data['% Out'].astype(float)

        # Round the '% Out' column to 3 decimal places
        formatted_data['% Out'] = formatted_data['% Out'].round(3)
        formatted_data.set_index('Date Reported', inplace=True)

        return formatted_data     
    

    def news(self, ticker:str):
        """
        Gets ticker news.

        Arguments:

        None
        """

        data = yf.Ticker(ticker=ticker).get_news()

        df=  pd.DataFrame(data)

      # Drop the 'uuid' column
        df = df.drop('uuid', axis=1)

        return df
    

    def atm_calls(self, ticker:str):
        """
        Gets at the money calls for a ticker.
 
        
        """

        calls = yf.Ticker(ticker)._download_options()

        call_options = calls['calls']


        df = pd.DataFrame(call_options)


        



        return df

    def atm_puts(self, ticker:str):
        """
        Gets At The Money puts for a ticker.
        
        """

        puts = yf.Ticker(ticker)._download_options()

        put_options = puts['puts']


        df = pd.DataFrame(put_options)


        



        return df




from asyncio import Lock
lock = Lock()
import numpy as np
import pandas as pd
from datetime import datetime
import asyncpg


async def batch_insert_dataframe(self, df, table_name, unique_columns, batch_size=250):
        """
        WORKS - Creates table - inserts data based on DTYPES.
        
        """
    
        async with lock:
            if not await self.table_exists(table_name):
                await self.create_table(df, table_name, unique_columns)
            
            # Debug: Print DataFrame columns before modifications
            #print("Initial DataFrame columns:", df.columns.tolist())
            
            df = df.copy()
            df.dropna(inplace=True)
            df['insertion_timestamp'] = [datetime.now() for _ in range(len(df))]

            # Debug: Print DataFrame columns after modifications
            #print("Modified DataFrame columns:", df.columns.tolist())
            
            records = df.to_records(index=False)
            data = list(records)


            async with self.pool.acquire() as connection:
                column_types = await connection.fetch(
                    f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'"
                )
                type_mapping = {col: next((item['data_type'] for item in column_types if item['column_name'] == col), None) for col in df.columns}

                async with connection.transaction():
                    insert_query = f"""
                    INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                    VALUES ({', '.join('$' + str(i) for i in range(1, len(df.columns) + 1))})
                    ON CONFLICT ({unique_columns})
                    DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                    """
            
                    batch_data = []
                    for record in data:
                        new_record = []
                        for col, val in zip(df.columns, record):
                
                            pg_type = type_mapping[col]

                            if val is None:
                                new_record.append(None)
                            elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))

            
                            elif isinstance(val, datetime):
                                new_record.append(pd.Timestamp(val).to_pydatetime())
                            elif pg_type in ['timestamp', 'timestamp without time zone', 'timestamp with time zone'] and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))  # Modified line
                            elif pg_type in ['double precision', 'real'] and not isinstance(val, str):
                                new_record.append(float(val))
                            elif isinstance(val, np.int64):  # Add this line to handle numpy.int64
                                new_record.append(int(val))
                            elif pg_type == 'integer' and not isinstance(val, int):
                                new_record.append(int(val))
                            else:
                                new_record.append(val)
                    
                        batch_data.append(new_record)

                        if len(batch_data) == batch_size:
                            try:
                                
                            
                                await connection.executemany(insert_query, batch_data)
                                batch_data.clear()
                            except Exception as e:
                                print(f"An error occurred while inserting the record: {e}")
                                await connection.execute('ROLLBACK')
                                raise

                if batch_data:  # Don't forget the last batch
    
                    try:

                        await connection.executemany(insert_query, batch_data)
                    except Exception as e:
                        print(f"An error occurred while inserting the record: {e}")
                        await connection.execute('ROLLBACK')
                        raise
async def save_to_history(self, df, main_table_name, history_table_name):
    # Assume the DataFrame `df` contains the records to be archived
    if not await self.table_exists(history_table_name):
        await self.create_table(df, history_table_name, None)

    df['archived_at'] = datetime.now()  # Add an 'archived_at' timestamp
    await self.batch_insert_dataframe(df, history_table_name, None)
async def table_exists(self, table_name):
    query = f"SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = '{table_name}');"

    async with self.pool.acquire() as conn:
        async with conn.transaction():
            exists = await conn.fetchval(query)
    return exists